#!/usr/bin/env python3
"""
Single-cell TelomereHunter2: Filtering, Sorting, and Summary for Barcodes

This script performs telomere analysis on single-cell sequencing data:
1. Runs TH2 bulk analysis with single-cell mode enabled
2. Processes barcodes above a read count threshold
3. Calculates per-barcode telomere content, TVR occurrences, and singleton variants
4. Outputs a comprehensive summary table with normalized metrics

The analysis follows the same approach as bulk TH2 but applied to each barcode independently.
"""
import sys
import os
import pandas as pd
import subprocess
import pysam
import numpy as np
from telomerehunter2.telomerehunter2_main import parse_command_line_arguments
from telomerehunter2.TVR_screen import find_prioritized_telomeric_indices
from telomerehunter2.utils import get_reverse_complement

# Default TVR hexamers if not specified in arguments
DEFAULT_TVR_HEXAMERS = [
    "TCAGGG", "TGAGGG", "TTGGGG", "TTCGGG", "TTTGGG",
    "ATAGGG", "CATGGG", "CTAGGG", "GTAGGG", "TAAGGG"
]

# BAM file types for telomere analysis
BAM_TYPES = ["intratelomeric", "junctionspanning", "subtelomeric", "intrachromosomal"]

# Canonical telomere repeat
CANONICAL_REPEAT = "TTAGGG"

# Default quality threshold for base calling
DEFAULT_QUALITY_THRESHOLD = 20


def count_reads_per_barcode_in_bamfiles(bam_files, barcodes):
    """
    Count reads for each barcode in each BAM file (intratelomeric, junctionspanning, subtelomeric, intrachromosomal).

    Args:
        bam_files: Dictionary mapping BAM types to file paths
        barcodes: List of barcodes to count reads for

    Returns:
        Dictionary mapping barcodes to counts per BAM type
    """
    barcode_counts = {bc: {bam_type: 0 for bam_type in bam_files} for bc in barcodes}
    total_barcodes = len(barcodes)

    print(f"Counting reads for {total_barcodes} barcodes across {len(bam_files)} BAM files")
    for bam_type, bam_path in bam_files.items():
        if not os.path.exists(bam_path):
            print(f"BAM file not found: {bam_path} (type: {bam_type})")
            continue
        print(f"Processing {bam_type} BAM file: {bam_path}")
        try:
            with pysam.AlignmentFile(bam_path, "rb") as bamfile:
                read_count = 0
                barcode_matched = 0
                for read in bamfile.fetch(until_eof=True):
                    read_count += 1
                    if read_count % 1000000 == 0:
                        print(f"Processed {read_count} reads, matched {barcode_matched} to barcodes")
                    bc = read.get_tag("CB") if read.has_tag("CB") else None
                    if bc in barcode_counts:
                        barcode_counts[bc][bam_type] += 1
                        barcode_matched += 1
                print(f"Completed {bam_type}: processed {read_count} reads, matched {barcode_matched} to barcodes")
        except Exception as e:
            print(f"Error processing {bam_type} BAM file: {str(e)}")
    return barcode_counts


def screen_bam_all_barcodes(intratel_bam_path, TVR_HEXAMERS, repeat_threshold_used=6, qual_threshold=DEFAULT_QUALITY_THRESHOLD):
    """
    Screen intratelomeric BAM for TVR hexamers and singletons for all barcodes in a single pass.

    Args:
        intratel_bam_path: Path to intratelomeric BAM file
        TVR_HEXAMERS: List of telomere variant repeat (TVR) hexamers to search for
        repeat_threshold_used: Minimum number of repeats required
        qual_threshold: Minimum base quality score required

    Returns:
        Tuple of (barcode_tvr_counts, barcode_singleton_counts, barcode_total_intratelomeric_bp)
    """
    barcode_tvr_counts = {}
    barcode_singleton_counts = {}
    barcode_total_intratelomeric_bp = {}

    if not os.path.exists(intratel_bam_path):
        print(f"Intratelomeric BAM not found: {intratel_bam_path}")
        return barcode_tvr_counts, barcode_singleton_counts, barcode_total_intratelomeric_bp

    print(f"Screening BAM file for TVRs and singletons: {intratel_bam_path}")
    print(f"Using repeat threshold: {repeat_threshold_used}, quality threshold: {qual_threshold}")
    with pysam.AlignmentFile(intratel_bam_path, "rb") as bamfile:
        read_count = 0
        for read in bamfile.fetch(until_eof=True):
            read_count += 1
            if read_count % 1000000 == 0:
                print(f"Processed {read_count} reads...")
            bc = read.get_tag("CB") if read.has_tag("CB") else None
            if not bc:
                continue

            seq = read.query_sequence
            qual = read.query_qualities

            # Initialize data structures for new barcodes
            if bc not in barcode_tvr_counts:
                barcode_tvr_counts[bc] = {tvr: 0 for tvr in TVR_HEXAMERS}
                barcode_singleton_counts[bc] = {tvr: 0 for tvr in TVR_HEXAMERS}
                barcode_total_intratelomeric_bp[bc] = 0

            barcode_total_intratelomeric_bp[bc] += len(seq)

            # Find telomeric pattern indices
            indices = find_prioritized_telomeric_indices(seq, telomere_pattern=r"(.{3})(GGG)")

            # Check reverse complement if not enough repeats found
            if len(indices) < repeat_threshold_used:
                seq = get_reverse_complement(seq)
                qual = read.query_qualities[::-1]  # Reverse quality scores too
                indices = find_prioritized_telomeric_indices(seq, telomere_pattern=r"(.{3})(GGG)")

            # Process each hexamer found
            for i in indices:
                if i + 6 > len(seq):  # Ensure we don't go past end of sequence
                    continue

                pattern_segment = seq[i:i+6]

                # Skip patterns with N or low quality
                if "N" in pattern_segment:
                    continue

                q_scores = qual[i:i+6]
                if any(q < qual_threshold for q in q_scores):
                    continue

                # Count pattern if it's in our list of TVRs
                if pattern_segment in barcode_tvr_counts[bc]:
                    barcode_tvr_counts[bc][pattern_segment] += 1

                    # Check for singleton (TVR flanked by canonical repeats)
                    if i >= 18 and i + 24 <= len(seq):
                        flank_left = seq[i-18:i]
                        flank_right = seq[i+6:i+24]
                        if flank_left == CANONICAL_REPEAT*3 and flank_right == CANONICAL_REPEAT*3:
                            barcode_singleton_counts[bc][pattern_segment] += 1

    print(f"Completed screening. Processed {read_count} reads across {len(barcode_tvr_counts)} barcodes")
    return barcode_tvr_counts, barcode_singleton_counts, barcode_total_intratelomeric_bp


def prepare_summary(barcodes_above, barcode_df, barcode_counts, barcode_tvr_counts, barcode_singleton_counts, barcode_total_intratelomeric_bp, args_th2, summary_df, TVR_HEXAMERS):
    """
    Prepare a summary dataframe for each barcode with telomere content, TVR occurrences, and singleton variants.

    Args:
        barcodes_above: List of barcodes above read count threshold
        barcode_df: DataFrame of barcode information
        barcode_counts: Dictionary mapping barcodes to read counts per BAM type
        barcode_tvr_counts: Dictionary mapping barcodes to TVR counts
        barcode_singleton_counts: Dictionary mapping barcodes to singleton counts
        barcode_total_intratelomeric_bp: Dictionary mapping barcodes to total intratelomeric base pair counts
        args_th2: Arguments from command line
        summary_df: Summary DataFrame from bulk TH2 run
        TVR_HEXAMERS: List of TVR hexamers to analyze

    Returns:
        List of dictionaries, each containing summary data for one barcode
    """
    total_reads_with_tel_gc = int(summary_df['total_reads_with_tel_gc'].values[0])
    total_reads = int(summary_df['total_reads'].values[0])
    pid = str(summary_df['PID'].values[0])
    read_length = str(summary_df['read_length'].values[0])
    repeat_threshold_set = str(summary_df['repeat_threshold_set'].values[0])
    repeat_threshold_used = str(summary_df['repeat_threshold_used'].values[0])
    gc_bins_for_correction = str(summary_df['gc_bins_for_correction'].values[0])

    print(f"Preparing summary for {len(barcodes_above)} barcodes")

    summary_rows = []
    for barcode in barcodes_above:
        # Get barcode-specific counts
        bc_total_reads = int(barcode_df[barcode_df['barcode'] == barcode]['read_count'].values[0])
        intratelomeric_reads = barcode_counts[barcode]["intratelomeric"]
        junctionspanning_reads = barcode_counts[barcode]["junctionspanning"]
        subtelomeric_reads = barcode_counts[barcode]["subtelomeric"]
        intrachromosomal_reads = barcode_counts[barcode]["intrachromosomal"]

        # Calculate telomere read count and content
        tel_read_count = intratelomeric_reads + junctionspanning_reads + subtelomeric_reads + intrachromosomal_reads
        bc_total_reads_with_tel_gc = (total_reads_with_tel_gc / total_reads) * bc_total_reads
        tel_content = (float(intratelomeric_reads) / float(bc_total_reads_with_tel_gc) * 1e6) if bc_total_reads_with_tel_gc > 0 else np.nan
        trpm = (float(tel_read_count) / float(bc_total_reads) * 1e6) if bc_total_reads > 0 else np.nan

        # Get TVR and singleton counts
        patterns = barcode_tvr_counts.get(barcode, {tvr: 0 for tvr in TVR_HEXAMERS})
        singleton_counts = barcode_singleton_counts.get(barcode, {tvr: 0 for tvr in TVR_HEXAMERS})
        summed_intratel_read_length = barcode_total_intratelomeric_bp.get(barcode, 0)

        # Calculate normalization values
        context_norm = {}
        context_norm_100bp = {}
        singleton_norm = {}

        for tvr in TVR_HEXAMERS:
            count = patterns.get(tvr, 0)
            context_norm[f"{tvr}_arbitrary_context_norm_by_intratel_reads"] = (count / intratelomeric_reads) if intratelomeric_reads > 0 else np.nan
            context_norm_100bp[f"{tvr}_arbitrary_context_per_100bp_intratel_read"] = (count * 100 / summed_intratel_read_length) if summed_intratel_read_length and summed_intratel_read_length > 0 else np.nan
            singleton_norm[f"{tvr}_singletons_norm_by_all_reads"] = (singleton_counts.get(tvr, 0) / bc_total_reads) if bc_total_reads > 0 else np.nan

        # Create summary row
        summary_row = {
            'PID': pid,
            'sample': barcode,
            'total_reads': bc_total_reads,
            'read_length': read_length,
            'repeat_threshold_set': repeat_threshold_set,
            'repeat_threshold_used': repeat_threshold_used,
            'intratelomeric_reads': intratelomeric_reads,
            'junctionspanning_reads': junctionspanning_reads,
            'subtelomeric_reads': subtelomeric_reads,
            'intrachromosomal_reads': intrachromosomal_reads,
            'tel_read_count': tel_read_count,
            'gc_bins_for_correction': gc_bins_for_correction,
            'total_reads_with_tel_gc': bc_total_reads_with_tel_gc,
            'tel_content': tel_content,
            'TRPM': trpm,
            **context_norm,
            **context_norm_100bp,
            **singleton_norm,
        }
        summary_rows.append(summary_row)

    print(f"Generated {len(summary_rows)} summary rows")
    return summary_rows


def main():
    """
    Main function to run single-cell TelomereHunter2 analysis.
    1. Run bulk TH2 analysis
    2. Process barcodes and count reads
    3. Screen for TVR hexamers and singletons
    4. Generate summary table
    """
    print("Running in TH2 single-cell mode: full file analysis...")

    # Parse command line arguments
    args_th2 = parse_command_line_arguments()

    # Ensure single-cell mode and disable plotting
    cli_args = list(sys.argv[1:])
    if '--plotNone' not in cli_args:
        cli_args.append('--plotNone')
    if '--singlecell_mode' not in cli_args:
        cli_args.append('--singlecell_mode')

    # Run bulk TH2 analysis first
    th2_cmd = [sys.executable, '-m', 'telomerehunter2.telomerehunter2_main', *cli_args]
    print(f"Running bulk TH2: {' '.join(map(str, th2_cmd))}")
    try:
        result = subprocess.run(th2_cmd, check=True)
        if result.returncode != 0:
            print("Bulk TH2 run failed. Exiting.")
            sys.exit(result.returncode)
    except subprocess.CalledProcessError as e:
        print(f"Bulk TH2 run failed with error: {str(e)}")
        sys.exit(1)

    # Load results from bulk run for SC analysis
    barcode_table = os.path.join(args_th2.outdir, f"{args_th2.pid}_barcode_counts.tsv")
    summary_file = os.path.join(args_th2.outdir, f"{args_th2.pid}_summary.tsv")

    if not os.path.exists(barcode_table):
        print(f"Barcode table not found: {barcode_table}")
        return
    if not os.path.exists(summary_file):
        print(f"Summary file not found: {summary_file}")
        return

    # Read result files
    try:
        barcode_df = pd.read_table(barcode_table)
        summary_df = pd.read_table(summary_file)
    except Exception as e:
        print(f"Error reading result files: {str(e)}")
        return

    # Check for required fields
    if 'total_reads_with_tel_gc' not in summary_df.columns:
        print("Required field 'total_reads_with_tel_gc' not found in summary file. Exiting.")
        sys.exit(1)

    # Filter barcodes by read count threshold
    barcodes_above = barcode_df[barcode_df['read_count'] >= args_th2.min_reads_per_barcode]['barcode'].tolist()
    print(f"Barcodes above threshold ({args_th2.min_reads_per_barcode} reads): {len(barcodes_above)}")
    if not barcodes_above:
        print("No barcodes above the threshold. Exiting.")
        return

    # Get BAM files and count reads per barcode
    bam_files_all = {
        bam_type: os.path.join(args_th2.outdir, f"tumor_TelomerCnt_{args_th2.pid}", f"{args_th2.pid}_filtered_{bam_type}.bam")
        for bam_type in BAM_TYPES
    }

    # Count reads per barcode in each BAM file
    barcode_counts = count_reads_per_barcode_in_bamfiles(bam_files=bam_files_all, barcodes=barcodes_above)

    # Use TVR hexamers from args or default list
    TVR_HEXAMERS = args_th2.TVRs_for_context if hasattr(args_th2, 'TVRs_for_context') else DEFAULT_TVR_HEXAMERS

    # Screen intratelomeric BAM for TVRs and singletons
    intratel_bam_path = os.path.join(
        args_th2.outdir,
        f"tumor_TelomerCnt_{args_th2.pid}",
        f"{args_th2.pid}_filtered_intratelomeric.bam"
    )

    barcode_tvr_counts, barcode_singleton_counts, barcode_total_intratelomeric_bp = screen_bam_all_barcodes(
        intratel_bam_path=intratel_bam_path,
        TVR_HEXAMERS=TVR_HEXAMERS,
        repeat_threshold_used=int(summary_df['repeat_threshold_used'].values[0]),
        qual_threshold=DEFAULT_QUALITY_THRESHOLD
    )

    # Prepare summary rows
    summary_rows = prepare_summary(
        barcodes_above,
        barcode_df,
        barcode_counts,
        barcode_tvr_counts,
        barcode_singleton_counts,
        barcode_total_intratelomeric_bp,
        args_th2,
        summary_df,
        TVR_HEXAMERS
    )

    print(f"Summary rows generated: {len(summary_rows)}")
    if summary_rows:
        # Get dynamic column names based on the TVR hexamers
        tvr_arbitrary = [f"{tvr}_arbitrary_context_norm_by_intratel_reads" for tvr in TVR_HEXAMERS]
        tvr_100bp = [f"{tvr}_arbitrary_context_per_100bp_intratel_read" for tvr in TVR_HEXAMERS]
        tvr_singletons = [f"{tvr}_singletons_norm_by_all_reads" for tvr in TVR_HEXAMERS]

        # Standard columns
        std_columns = [
            'PID', 'sample', 'total_reads', 'read_length', 'repeat_threshold_set', 'repeat_threshold_used',
            'intratelomeric_reads', 'junctionspanning_reads', 'subtelomeric_reads', 'intrachromosomal_reads',
            'tel_read_count', 'gc_bins_for_correction', 'total_reads_with_tel_gc', 'tel_content', 'TRPM'
        ]

        # All columns
        columns = std_columns + tvr_arbitrary + tvr_100bp + tvr_singletons

        # Create DataFrame and select columns
        summary_df = pd.DataFrame(summary_rows)

        # Ensure all expected columns exist
        for col in columns:
            if col not in summary_df.columns:
                print(f"Expected column '{col}' not found in results. Adding empty column.")
                summary_df[col] = np.nan

        summary_final = summary_df[columns]

        # Write to file
        summary_path = os.path.join(args_th2.outdir, f"{args_th2.pid}_sc_summary.tsv")
        summary_final.to_csv(summary_path, sep="\t", index=False)
        print(f"SC summary file written to: {summary_path}")
    else:
        print("No barcode summaries generated.")
    print("Completed barcode-specific analysis and summary writing.")


if __name__ == "__main__":
    main()



#!/usr/bin/python

# Copyright 2024 Ferdinand Popp, Lina Sieverling, Philip Ginsbach, Lars Feuerbach

# This file is part of TelomereHunter2.

# TelomereHunter2 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# TelomereHunter2 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with TelomereHunter2. If not, see <http://www.gnu.org/licenses/>.


import os
import re

import pandas as pd
import pysam

from telomerehunter2.utils import get_reverse_complement

#####################################################################################
###
### get the context of telomere variant repeats (TVRs)
###
#####################################################################################


def dictionary_to_table(dictionary, outfile_path, cutoff):
    try:
        # write sorted dictionary as a table to file
        df = pd.DataFrame.from_dict(dictionary, orient="index").reset_index()
        df.columns = ["Bases", "Count"]
        df = df.sort_values(by="Count", ascending=False)

        # add per all instances
        df["Percent"] = (df["Count"] / df["Count"].sum()) * 100

        # remove ones below cutoff
        df = df[df["Count"] >= cutoff]

        df = df.convert_dtypes()
        df.to_csv(outfile_path, sep="\t", index=False)

    except (ValueError, KeyError):
        # Handle the case where the dictionary is empty
        df = pd.DataFrame(columns=["Bases", "Count", "Percent"])
        df.to_csv(outfile_path, sep="\t", index=False)


def tvr_context(
    main_path,
    pid,
    sample,
    pattern,
    min_base_quality=20,
    context_before=18,
    context_after=18,
    telomere_pattern="GGG",
    repeat_threshold_set=6,
    cutoff=0,
    tel_file="filtered_intratelomeric",
):
    """
    # main_path: Path to TelomereHunter results
    # pattern: Pattern for which to get context
    # min_base_quality: Minimum base quality required for pattern
    # context_before: Number of bases before start of pattern
    # context_after: Number of bases after end of pattern
    # telomere_pattern: Pattern with which to identify G-rich reads
    # repeat_threshold_set: int number of telomeric reads on read threshold
    # cutoff: Count cutoff for displaying neighborhood in output table
    # tel_file: the telomere file in which to search for TVRs (filtered, filtered_intratelomeric, ...)
    """
    outdir = os.path.join(main_path, "TVR_context")

    os.makedirs(outdir, exist_ok=True)

    # Get BAM reads
    bam_file_path = os.path.join(main_path, f"{pid}_{tel_file}.bam")
    bamfile = pysam.AlignmentFile(bam_file_path, "rb")

    # directories for counting of patterns in neighborhood
    neighborhood_before = {}
    neighborhood_after = {}
    neighborhood = {}

    for read in bamfile.fetch(until_eof=True):

        seq = read.query_sequence
        qual = read.query_qualities

        # find GGG counted once for reverse complement decision
        indices_telomeric = [m.start() for m in re.finditer(telomere_pattern, seq)]

        # get reverse complement of sequence if the telomeric pattern was not found often enough
        if len(indices_telomeric) < repeat_threshold_set:
            seq = get_reverse_complement(seq)
            qual = read.query_qualities[::-1]

        indices_pattern = [m.start() for m in re.finditer(rf"(?={pattern})", seq)]

        for i in indices_pattern:

            bases_before = ""
            bases_after = ""

            # get base qualities of pattern
            base_qualities = [qual[i + j] for j in range(6)]
            if any(q < min_base_quality for q in base_qualities):
                continue

            # todo check if keyerror needs to be caught
            if i - context_before >= 0:
                bases_before = seq[i - context_before : i]
                neighborhood_before[bases_before] = (
                    neighborhood_before.get(bases_before, 0) + 1
                )

            if i + len(pattern) + context_after <= len(seq):
                bases_after = seq[i + len(pattern) : i + len(pattern) + context_after]
                neighborhood_after[bases_after] = (
                    neighborhood_after.get(bases_after, 0) + 1
                )

            if bases_before and bases_after:
                bases_all = f"{bases_before}-{pattern}-{bases_after}"
                neighborhood[bases_all] = neighborhood.get(bases_all, 0) + 1

    dictionary_to_table(
        neighborhood_before,
        os.path.join(
            outdir,
            f"{pid}_{sample}_{pattern}_{context_before}bp_neighborhood_before.tsv",
        ),
        cutoff=cutoff,
    )
    dictionary_to_table(
        neighborhood_after,
        os.path.join(
            outdir, f"{pid}_{sample}_{pattern}_{context_after}bp_neighborhood_after.tsv"
        ),
        cutoff=cutoff,
    )
    dictionary_to_table(
        neighborhood,
        os.path.join(
            outdir,
            f"{pid}_{sample}_{pattern}_{context_before}bp_{context_after}bp_neighborhood.tsv",
        ),
        cutoff=cutoff,
    )



#!/usr/bin/python

# Copyright 2024 Ferdinand Popp, Lina Sieverling, Philip Ginsbach, Lars Feuerbach

# This file is part of TelomereHunter2.

# TelomereHunter2 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# TelomereHunter2 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with TelomereHunter2. If not, see <http://www.gnu.org/licenses/>.


import os

import pysam

from telomerehunter2.utils import (get_band_info, get_reverse_complement,
                                   measure_time)


def make_band_and_spectrum_lists(band_file, patterns):
    """
    Create band and spectrum lists based on the information from a banding file.

    Parameters:
    - band_file (str): Path to the banding file.
    - patterns (list): List of patterns for spectrum initialization.

    Returns:
    - bands_list (dict): Dictionary containing band information for each chromosome.
    - spectrum_list (dict): Dictionary containing spectrum information for each chromosome and band.
    - chromosome_list (list): List of chromosome names.

    """
    # Get chromosome information
    _, chromosome_list = get_band_info(band_file)

    # Initialize bands_list with unmapped entry
    bands_list = {"unmapped": {"band_name": ["unmapped"], "end": [0]}}
    bands_list.update({chr: {"band_name": [], "end": []} for chr in chromosome_list})

    # Initialize spectrum_list dictionary
    spectrum_list = {
        chr: {
            band_name: {
                "other": 0.0,
                "reads_with_pattern": 0,
                **{pattern: 0 for pattern in patterns},
            }
            for band_name in ["junction1", "junction2"]
        }
        for chr in chromosome_list
    }
    spectrum_list["unmapped"] = {
        "unmapped": {
            "other": 0.0,
            "reads_with_pattern": 0,
            **{pattern: 0 for pattern in patterns},
        }
    }

    # Process banding file
    with open(band_file, "r") as band_file:
        for line in band_file:
            try:
                # Parse banding file line
                chrom, _, end, band_name, *additional_items = line.strip().split()
                chrom = chrom[3:] if chrom.startswith("chr") else chrom

                # Update bands_list with band information
                bands_list[chrom]["band_name"].append(band_name)
                bands_list[chrom]["end"].append(int(end))

                # Initialize spectrum_list for the current band
                spectrum_list[chrom][band_name] = {
                    "other": 0.0,
                    "reads_with_pattern": 0,
                    **{pattern: 0 for pattern in patterns},
                }

            except Exception as e:
                print(f"Invalid line in banding file: '{line}'")

    return bands_list, spectrum_list, chromosome_list


def write_spectrum_file(
    out_dir, pid, chromosome_list, bands_list, spectrum_list, patterns
):
    spectrum_file_path = os.path.join(out_dir, f"{pid}_spectrum.tsv")

    if not spectrum_list:
        spectrum_list = get_default_spectrum_list(patterns, spectrum_list)

    with open(spectrum_file_path, "w") as spectrum_file:
        spectrum_file.write("chr\tband\treads_with_pattern")

        for pattern in patterns:
            spectrum_file.write("\t" + pattern)

        spectrum_file.write("\tother\n")

        for chromosome in chromosome_list:
            for band in (
                ["junction1"] + bands_list[chromosome]["band_name"] + ["junction2"]
            ):
                spectrum_file.write(
                    f"{chromosome}\t{band}\t{spectrum_list[chromosome][band]['reads_with_pattern']}"
                )

                for pattern in patterns:
                    spectrum_file.write(f"\t{spectrum_list[chromosome][band][pattern]}")

                spectrum_file.write(
                    f"\t{int(round(spectrum_list[chromosome][band]['other']))}\n"
                )

        spectrum_file.write(
            f"unmapped\tunmapped\t{spectrum_list['unmapped']['unmapped']['reads_with_pattern']}"
        )

        for pattern in patterns:
            spectrum_file.write(f"\t{spectrum_list['unmapped']['unmapped'][pattern]}")

        spectrum_file.write(
            f"\t{int(round(spectrum_list['unmapped']['unmapped']['other']))}\n"
        )


def get_default_spectrum_list(patterns, spectrum_list):
    spectrum_list = {
        "unmapped": {
            "unmapped": {
                "reads_with_pattern": 0,
                "other": 0,
                **{pattern: 0 for pattern in patterns},
            }
        }
    }
    return spectrum_list


def close_file_handles(
    bamfile,
    intratelomeric_file,
    junctionspanning_file,
    subtelomeric_file,
    intrachromosomal_file,
):
    bamfile.close()
    intratelomeric_file.close()
    junctionspanning_file.close()
    subtelomeric_file.close()
    intrachromosomal_file.close()


# function for checking reads without mate and sorting
def read_check(read, references, bands_list, chromosome_list, mapq_threshold):
    """
    Check properties of a read and return relevant information.

    Parameters:
    - read: object, BAM read object.
    - references: list, list of reference sequences.
    - bands_list: dict, dictionary containing band information for each chromosome.
    - chromosome_list: list, list of valid chromosome names.
    - mapq_threshold: int, minimum mapping quality threshold.

    Returns:
    Tuple containing information about the read:
    (chromosome, band, read_is_unmapped, read_junctionspanning, read_junction)
    """

    # get reference
    read_tid = read.tid
    read_ref = references[read_tid].replace("chr", "") if read_tid != -1 else ""

    read_junctionspanning = False
    read_junction = ""

    # check if read is considered unmapped
    if (
        read.is_unmapped
        or read_ref not in chromosome_list
        or read.mapping_quality < mapq_threshold
    ):
        read_is_unmapped = True
        chromosome, band = "unmapped", "unmapped"
    # map read
    else:
        read_is_unmapped = False
        chromosome = read_ref

        # find band
        i = 0
        while read.pos > bands_list[chromosome]["end"][i] and i < (
            len(bands_list[chromosome]["end"]) - 1
        ):
            i += 1
        else:
            band = bands_list[chromosome]["band_name"][i]

            # check if read is mapped to first or last chromosome band
            if i == 0:
                read_junctionspanning = True
                read_junction = "junction1"
            elif i == (len(bands_list[chromosome]["end"]) - 1):
                read_junctionspanning = True
                read_junction = "junction2"

    return chromosome, band, read_is_unmapped, read_junctionspanning, read_junction


# function for sorting reads without a mate into correct fraction and add counts to spectrum list
def sort_reads_without_mate(
    read,
    references,
    bands_list,
    chromosome_list,
    mapq_threshold,
    spectrum_list,
    patterns,
    intratelomeric_file,
    subtelomeric_file,
    intrachromosomal_file,
):
    """
    Sort reads without a mate into the correct fraction and update counts.

    Parameters:
    - read: object, BAM read object.
    - references: list, list of reference sequences.
    - bands_list: dict, dictionary containing band information for each chromosome.
    - chromosome_list: list, list of valid chromosome names.
    - mapq_threshold: int, minimum mapping quality threshold.
    - spectrum_list: dict, dictionary to store counts of different patterns for each chromosome and band.
    - patterns: list, with base patterns
    - intratelomeric_file: file, output file for intratelomeric reads.
    - subtelomeric_file: file, output file for subtelomeric reads.
    - intrachromosomal_file: file, output file for intrachromosomal reads.
    """

    # do read check
    chromosome, band, read_is_unmapped, read_junctionspanning, read_junction = (
        read_check(read, references, bands_list, chromosome_list, mapq_threshold)
    )

    # write read to correct fraction
    if read_is_unmapped:
        intratelomeric_file.write(read)
    elif read_junctionspanning:
        subtelomeric_file.write(read)
    else:
        intrachromosomal_file.write(read)

    # add read counts to spectrum list
    spectrum_list[chromosome][band]["reads_with_pattern"] += 1

    read_total_pattern_count = 0

    for pattern in patterns:
        pattern_count = read.query_sequence.count(pattern)
        spectrum_list[chromosome][band][pattern] += pattern_count
        read_total_pattern_count += pattern_count

    # get count of other e.g. hexamers found
    k_mer_length = len(patterns[0])
    spectrum_list[chromosome][band]["other"] += (
        float(len(read.query_sequence)) / k_mer_length
    ) - read_total_pattern_count


# function for writing reads with mates to fraction and adding counts to spectrum list
def sort_read_with_mate(read, fraction_file, chromosome, band, spectrum_list, patterns):
    """
    Sort reads with a mate into the correct fraction and update counts.

    Parameters:
    - read: object, BAM read object.
    - fraction_file: file, output file for the specific fraction.
    - chromosome: str, chromosome name.
    - band: str, band name.
    - spectrum_list: dict, dictionary to store counts of different patterns for each chromosome and band.
    - patterns: list of target base patterns
    """

    fraction_file.write(read)

    spectrum_list[chromosome][band]["reads_with_pattern"] += 1

    read_total_pattern_count = 0

    # count occurrence of pattern types
    for pattern in patterns:
        pattern_count = read.query_sequence.count(pattern)
        spectrum_list[chromosome][band][pattern] += pattern_count
        read_total_pattern_count += pattern_count

    # get count of other e.g. hexamers found
    k_mer_length = len(patterns[0])
    spectrum_list[chromosome][band]["other"] += (
        float(len(read.query_sequence)) / k_mer_length
    ) - read_total_pattern_count


def process_reads(
    bamfile,
    chromosome_list,
    bands_list,
    spectrum_list,
    patterns,
    mapq_threshold,
    intratelomeric_file,
    junctionspanning_file,
    subtelomeric_file,
    intrachromosomal_file,
):
    """
    Process paired-end reads and categorize them into different fractions based on mapping characteristics.

    Parameters:
    - bamfile: iterator, iterator over BAM file containing paired-end reads.
    - chromosome_list: list, list of valid chromosome names.
    - bands_list: dict, dictionary containing band information for each chromosome.
    - spectrum_list: dict, dictionary to store counts of different patterns for each chromosome and band.
    - patterns: list, target base sequences
    - mapq_threshold: int, minimum mapping quality threshold.
    - intratelomeric_file: file, output file for intratelomeric reads.
    - junctionspanning_file: file, output file for junction spanning reads.
    - subtelomeric_file: file, output file for subtelomeric reads.
    - intrachromosomal_file: file, output file for intrachromosomal reads.

    Returns:
    Tuple: chromosome_list, bands_list, spectrum_list
    """

    break_flag = False
    references = bamfile.references

    while True:
        try:
            read1 = next(bamfile)
        except StopIteration:
            break

        try:
            read2 = next(bamfile)
        except StopIteration:
            sort_reads_without_mate(
                read1,
                references,
                bands_list,
                chromosome_list,
                mapq_threshold,
                spectrum_list,
                patterns,
                intratelomeric_file,
                subtelomeric_file,
                intrachromosomal_file,
            )
            break

        read1_first_flag = True

        # while not mates
        while read1.query_name != read2.query_name:
            if read1_first_flag:
                sort_reads_without_mate(
                    read1,
                    references,
                    bands_list,
                    chromosome_list,
                    mapq_threshold,
                    spectrum_list,
                    patterns,
                    intratelomeric_file,
                    subtelomeric_file,
                    intrachromosomal_file,
                )
                try:
                    read1 = next(bamfile)
                except StopIteration:
                    sort_reads_without_mate(
                        read2,
                        references,
                        bands_list,
                        chromosome_list,
                        mapq_threshold,
                        spectrum_list,
                        patterns,
                        intratelomeric_file,
                        subtelomeric_file,
                        intrachromosomal_file,
                    )
                    break_flag = True
                    break
                read1_first_flag = False
            else:
                sort_reads_without_mate(
                    read2,
                    references,
                    bands_list,
                    chromosome_list,
                    mapq_threshold,
                    spectrum_list,
                    patterns,
                    intratelomeric_file,
                    subtelomeric_file,
                    intrachromosomal_file,
                )
                try:
                    read2 = next(bamfile)
                except StopIteration:
                    sort_reads_without_mate(
                        read1,
                        references,
                        bands_list,
                        chromosome_list,
                        mapq_threshold,
                        spectrum_list,
                        patterns,
                        intratelomeric_file,
                        subtelomeric_file,
                        intrachromosomal_file,
                    )
                    break_flag = True
                    break
                read1_first_flag = True

        if break_flag:
            break

        # Complete Mates
        (
            read1_chromosome,
            read1_band,
            read1_is_unmapped,
            read1_junctionspanning,
            read1_junction,
        ) = read_check(
            read1, bamfile.references, bands_list, chromosome_list, mapq_threshold
        )
        (
            read2_chromosome,
            read2_band,
            read2_is_unmapped,
            read2_junctionspanning,
            read2_junction,
        ) = read_check(
            read2, bamfile.references, bands_list, chromosome_list, mapq_threshold
        )

        # INTRATELOMERIC: both mates are unmapped
        if read1_is_unmapped and read2_is_unmapped:

            sort_read_with_mate(
                read=read1,
                fraction_file=intratelomeric_file,
                chromosome="unmapped",
                band="unmapped",
                spectrum_list=spectrum_list,
                patterns=patterns,
            )
            sort_read_with_mate(
                read=read2,
                fraction_file=intratelomeric_file,
                chromosome="unmapped",
                band="unmapped",
                spectrum_list=spectrum_list,
                patterns=patterns,
            )

        # JUNCTION SPANNING: one mate is unmapped and the other mate is mapped to first or last band of chromosome
        elif (read1_is_unmapped and read2_junctionspanning) or (
            read2_is_unmapped and read1_junctionspanning
        ):

            if read1_is_unmapped:
                chromosome = read2_chromosome
                band = read2_junction
            else:
                chromosome = read1_chromosome
                band = read1_junction

            sort_read_with_mate(
                read=read1,
                fraction_file=junctionspanning_file,
                chromosome=chromosome,
                band=band,
                spectrum_list=spectrum_list,
                patterns=patterns,
            )
            sort_read_with_mate(
                read=read2,
                fraction_file=junctionspanning_file,
                chromosome=chromosome,
                band=band,
                spectrum_list=spectrum_list,
                patterns=patterns,
            )

        # SUBTELOMERIC: both mates are mapped to first or last band or chromosome (and have similar mapping positions)
        elif read1_junctionspanning and read2_junctionspanning:

            sort_read_with_mate(
                read=read1,
                fraction_file=subtelomeric_file,
                chromosome=read1_chromosome,
                band=read1_band,
                spectrum_list=spectrum_list,
                patterns=patterns,
            )
            sort_read_with_mate(
                read=read2,
                fraction_file=subtelomeric_file,
                chromosome=read2_chromosome,
                band=read2_band,
                spectrum_list=spectrum_list,
                patterns=patterns,
            )

        # SUBTELOMERIC/INTRACHROMOSOMAL: one read is subtelomeric, the other is intra-chromosomal => count and sort individually
        elif read1_junctionspanning or read2_junctionspanning:

            if read1_junctionspanning:
                read1_file = subtelomeric_file
                read2_file = intrachromosomal_file
            else:
                read2_file = subtelomeric_file
                read1_file = intrachromosomal_file

            sort_read_with_mate(
                read=read1,
                fraction_file=read1_file,
                chromosome=read1_chromosome,
                band=read1_band,
                spectrum_list=spectrum_list,
                patterns=patterns,
            )
            sort_read_with_mate(
                read=read2,
                fraction_file=read2_file,
                chromosome=read2_chromosome,
                band=read2_band,
                spectrum_list=spectrum_list,
                patterns=patterns,
            )

        # INTRACHROMOSOMAL: one read is intra-chromosomal, the other is unmapped => count both to position of mapped read
        elif (read1_is_unmapped and not read2_junctionspanning) or (
            read2_is_unmapped and not read1_junctionspanning
        ):

            if read1_is_unmapped:
                chromosome = read2_chromosome
                band = read2_band
            else:
                chromosome = read1_chromosome
                band = read1_band

            sort_read_with_mate(
                read=read1,
                fraction_file=intrachromosomal_file,
                chromosome=chromosome,
                band=band,
                spectrum_list=spectrum_list,
                patterns=patterns,
            )
            sort_read_with_mate(
                read=read2,
                fraction_file=intrachromosomal_file,
                chromosome=chromosome,
                band=band,
                spectrum_list=spectrum_list,
                patterns=patterns,
            )

        # INTRACHROMOSOMAL: both mapped
        else:

            sort_read_with_mate(
                read=read1,
                fraction_file=intrachromosomal_file,
                chromosome=read1_chromosome,
                band=read1_band,
                spectrum_list=spectrum_list,
                patterns=patterns,
            )
            sort_read_with_mate(
                read=read2,
                fraction_file=intrachromosomal_file,
                chromosome=read2_chromosome,
                band=read2_band,
                spectrum_list=spectrum_list,
                patterns=patterns,
            )

    return chromosome_list, bands_list, spectrum_list


def write_spectrum_file_no_banding(out_dir, pid, spectrum_list, patterns):
    spectrum_file_path = os.path.join(out_dir, f"{pid}_spectrum.tsv")

    if not spectrum_list:
        spectrum_list = get_default_spectrum_list(patterns, spectrum_list)

    with open(spectrum_file_path, "w") as spectrum_file:
        spectrum_file.write("chr\tband\treads_with_pattern")

        for pattern in patterns:
            spectrum_file.write("\t" + pattern)

        spectrum_file.write("\tother\n")

        for ref in spectrum_list:
            for band in spectrum_list[ref]:
                spectrum_file.write(
                    f"{ref}\t{band}\t{spectrum_list[ref][band]['reads_with_pattern']}"
                )

                for pattern in patterns:
                    spectrum_file.write(f"\t{spectrum_list[ref][band][pattern]}")

                spectrum_file.write(
                    f"\t{int(round(spectrum_list[ref][band]['other']))}\n"
                )


def count_patterns_in_read_no_banding(read, mapq_threshold, patterns, spectrum_list):
    """
    Counts the occurrences of specified patterns in a read and updates the spectrum list.

    Args:
        read (pysam.AlignedSegment): Aligned read object.
        mapq_threshold (int): Minimum mapping quality threshold for considering a read mapped.
        patterns (list): List of patterns to count occurrences of.
        spectrum_list (dict): Dictionary containing counts of patterns for different reference sequences and bands.

    Returns:
        dict: Updated spectrum list after counting occurrences of patterns in the read.
    """

    # Unmapped or low qual reads should be "unmapped"
    is_unmapped = read.is_unmapped or read.mapping_quality < mapq_threshold
    ref = "unmapped" if is_unmapped else read.reference_name
    band = "unmapped" if is_unmapped else "_"

    read_total_pattern_count = 0

    # Intialize the dict counts (patterns + other) to store other hexamers counts
    if ref not in spectrum_list:
        spectrum_list[ref] = {band: {"other": 0, "reads_with_pattern": 0}}

    # count occurrence of pattern types
    for pattern in patterns:
        if pattern not in spectrum_list[ref][band]:
            spectrum_list[ref][band][pattern] = 0

        pattern_count = read.query_sequence.count(pattern)
        spectrum_list[ref][band][pattern] += pattern_count
        read_total_pattern_count += pattern_count

    # get count of other e.g. hexamers found
    k_mer_length = len(patterns[0])
    spectrum_list[ref][band]["other"] += (
        float(len(read.query_sequence)) / k_mer_length
    ) - read_total_pattern_count

    # count in general the read with count
    spectrum_list[ref][band]["reads_with_pattern"] += 1

    return spectrum_list


def sort_read_without_mate_no_banding(
    read,
    mapq_threshold,
    patterns,
    intratelomeric_file,
    intrachromosomal_file,
    spectrum_list,
):
    is_mapped_read = read.is_mapped and read.mapping_quality >= mapq_threshold
    # Categorize reads and write them to respective files
    if not is_mapped_read:
        # not mapped -> telo
        intratelomeric_file.write(read)
    else:
        # mapped -> chr
        intrachromosomal_file.write(read)

    # Count patterns in read sequences
    spectrum_list = count_patterns_in_read_no_banding(
        read, mapq_threshold, patterns, spectrum_list
    )

    return spectrum_list


def process_reads_no_banding(
    bamfile, patterns, mapq_threshold, intratelomeric_file, intrachromosomal_file
):
    break_flag = False
    # dict with levels: reference_name, band, count
    spectrum_list = {}

    while True:
        try:
            read1 = next(bamfile)
        except StopIteration:
            break

        try:
            read2 = next(bamfile)
        except StopIteration:
            sort_read_without_mate_no_banding(
                read1,
                mapq_threshold,
                patterns,
                intratelomeric_file,
                intrachromosomal_file,
                spectrum_list,
            )
            break

        read1_first_flag = True

        # while not mates iterate over read 1 and 2 step by step
        while read1.query_name != read2.query_name:
            # exchange read 1
            if read1_first_flag:
                sort_read_without_mate_no_banding(
                    read1,
                    mapq_threshold,
                    patterns,
                    intratelomeric_file,
                    intrachromosomal_file,
                    spectrum_list,
                )
                try:
                    read1 = next(bamfile)
                except StopIteration:
                    sort_read_without_mate_no_banding(
                        read2,
                        mapq_threshold,
                        patterns,
                        intratelomeric_file,
                        intrachromosomal_file,
                        spectrum_list,
                    )
                    break_flag = True
                    break
                read1_first_flag = False
            # exchange read 2
            else:
                sort_read_without_mate_no_banding(
                    read2,
                    mapq_threshold,
                    patterns,
                    intratelomeric_file,
                    intrachromosomal_file,
                    spectrum_list,
                )
                try:
                    read2 = next(bamfile)
                except StopIteration:
                    sort_read_without_mate_no_banding(
                        read1,
                        mapq_threshold,
                        patterns,
                        intratelomeric_file,
                        intrachromosomal_file,
                        spectrum_list,
                    )
                    break_flag = True
                    break
                read1_first_flag = True

        if break_flag:
            break

        # True mates

        # Check mapping quality and determine if reads are mapped
        is_mapped_read1 = read1.is_mapped and read1.mapping_quality >= mapq_threshold
        is_mapped_read2 = read2.is_mapped and read2.mapping_quality >= mapq_threshold

        # Categorize reads and write them to respective files
        if not is_mapped_read1 and not is_mapped_read2:
            # both not mapped -> telomeric
            intratelomeric_file.write(read1)
            intratelomeric_file.write(read2)
        else:
            # one or both mapped -> chromsomal
            intrachromosomal_file.write(read1)
            intrachromosomal_file.write(read2)

        # Count patterns in read sequences
        spectrum_list = count_patterns_in_read_no_banding(
            read1, mapq_threshold, patterns, spectrum_list
        )
        spectrum_list = count_patterns_in_read_no_banding(
            read2, mapq_threshold, patterns, spectrum_list
        )

    return spectrum_list


@measure_time
def sort_telomere_reads(input_dir, band_file, out_dir, pid, mapq_threshold, repeats):
    """
    Sorts the reads from an input BAM file (needs to be name sorted!) containing telomere reads into 4 different fractions:
    intratelomeric, junction-spanning, subtelomeric, intrachromosomal. Additionally, a table containing the number of
    telomere reads and the pattern occurrences per chromosome band is written.

    Parameters:
        input_dir (str): Path to the directory containing input BAM file.
        band_file (str): Path to the file containing chromosome band information. Pass None if not available.
        out_dir (str): Path to the directory where output files will be written.
        pid (str): Process ID or identifier for the output files.
        mapq_threshold (int): Mapping quality threshold for filtering reads.
        repeats (list): List of telomere repeat sequences.

    Returns:
        None
    """

    #####################
    ### get patterns  ###
    #####################

    # Extract patterns and reverse complements
    patterns = []
    for repeat in repeats:
        patterns.append(repeat)
        patterns.append(get_reverse_complement(repeat))

    # Open input BAM file for reading
    bamfile = pysam.AlignmentFile(
        os.path.join(input_dir, f"{pid}_filtered_name_sorted.bam"), "rb"
    )

    # Open filtered files for writing
    intratelomeric_file = pysam.AlignmentFile(
        os.path.join(out_dir, f"{pid}_filtered_intratelomeric.bam"),
        "wb",
        template=bamfile,
    )
    junctionspanning_file = pysam.AlignmentFile(
        os.path.join(out_dir, f"{pid}_filtered_junctionspanning.bam"),
        "wb",
        template=bamfile,
    )
    subtelomeric_file = pysam.AlignmentFile(
        os.path.join(out_dir, f"{pid}_filtered_subtelomeric.bam"),
        "wb",
        template=bamfile,
    )
    intrachromosomal_file = pysam.AlignmentFile(
        os.path.join(out_dir, f"{pid}_filtered_intrachromosomal.bam"),
        "wb",
        template=bamfile,
    )

    #################################################################
    ### go through name sorted BAM file containing telomere reads ###
    ### => always look at mate pairs (if possible)                ###
    #################################################################

    # run classic version
    if band_file:
        # Make band and spectrum lists
        bands_list, spectrum_list, chromosome_list = make_band_and_spectrum_lists(
            band_file, patterns
        )

        # Process reads writing to the files
        chromosome_list, bands_list, spectrum_list = process_reads(
            bamfile,
            chromosome_list,
            bands_list,
            spectrum_list,
            patterns,
            mapq_threshold,
            intratelomeric_file,
            junctionspanning_file,
            subtelomeric_file,
            intrachromosomal_file,
        )

        # Write spectrum file
        write_spectrum_file(
            out_dir, pid, chromosome_list, bands_list, spectrum_list, patterns
        )

    # run without banding information (fast)
    else:
        # Process reads writing to the files, taking references from reads
        spectrum_list = process_reads_no_banding(
            bamfile,
            patterns,
            mapq_threshold,
            intratelomeric_file,
            intrachromosomal_file,
        )

        # Write spectrum file slim version
        write_spectrum_file_no_banding(out_dir, pid, spectrum_list, patterns)

    # Close file handles
    close_file_handles(
        bamfile,
        intratelomeric_file,
        junctionspanning_file,
        subtelomeric_file,
        intrachromosomal_file,
    )



#!/usr/bin/python

# Copyright 2024 Ferdinand Popp, Lina Sieverling, Philip Ginsbach, Lars Feuerbach

# This file is part of TelomereHunter2.

# TelomereHunter2 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# TelomereHunter2 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with TelomereHunter2. If not, see <http://www.gnu.org/licenses/>.

import os
import re

import pysam

from telomerehunter2.utils import get_reverse_complement, measure_time

###############################################################################################
### get the distribution of telomeric repeats per intratelomeric read in the input BAM file ###
###############################################################################################


@measure_time
def repeat_frequency_intratelomeric(input_path, out_dir, pid, repeats):
    ################################################
    ### get patterns and make regular expression ###
    ################################################

    # Create a regular expression pattern for forward and reverse repeats
    patterns_regex_forward = "|".join(repeats)
    patterns_regex_reverse = "|".join(
        get_reverse_complement(repeat) for repeat in repeats
    )

    #########################
    ### open file handles ###
    #########################

    # open input bam_file for reading
    bamfile = pysam.AlignmentFile(
        os.path.join(input_path, f"{pid}_filtered_intratelomeric.bam"), "rb"
    )

    ##################################
    ### initialize frequency table ###
    ##################################

    frequency_table = {repeats: 0 for repeats in range(0, 2 + 1)}

    ######################################
    ### loop through filtered BAM file ###
    ######################################

    for read in bamfile.fetch(until_eof=True):
        sequence = read.query_sequence
        number_repeats_forward = len(re.findall(patterns_regex_forward, sequence))
        number_repeats_reverse = len(re.findall(patterns_regex_reverse, sequence))
        number_repeats = max(number_repeats_forward, number_repeats_reverse)

        # Increment count for the number of repeats or set to 1 if the key is not present
        try:
            frequency_table[number_repeats] += 1
        except KeyError:
            frequency_table[number_repeats] = 1

    ##################################
    ### write frequency table file ###
    ##################################

    frequency_table_filepath = os.path.join(
        out_dir, f"{pid}_repeat_frequency_per_intratelomeric_read.tsv"
    )
    with open(frequency_table_filepath, "w") as frequency_table_file:
        # Header
        frequency_table_file.write("number_repeats\tcount\n")

        # Write line for each frequency
        for frequency, count in sorted(frequency_table.items()):
            frequency_table_file.write(f"{frequency}\t{count}\n")

    ##########################
    ### close file handles ###
    ##########################

    bamfile.close()



import os
import numpy as np
import pandas as pd
import pysam
from telomerehunter2 import filter_telomere_reads
from telomerehunter2.utils import get_read_lengths_and_repeat_thresholds

def run_fast_mode(args):
    out_dir = os.path.join(args.parent_outdir, args.pid)
    os.makedirs(out_dir, exist_ok=True)
    band_file = getattr(args, 'banding_file', None)

    tumor_summary = None
    control_summary = None
    if args.tumor_bam:
        tumor_summary = process_fast_mode_sample(args, args.tumor_bam, "tumor", out_dir, band_file)
    if args.control_bam:
        control_summary = process_fast_mode_sample(args, args.control_bam, "control", out_dir, band_file)
    write_fast_mode_combined_summary(args, out_dir, tumor_summary, control_summary)


def process_fast_mode_sample(args, bam_path, sample_name, out_dir, band_file):
    sample_out_dir = os.path.join(out_dir, f"{sample_name}_TelomerCnt_{args.pid}")
    os.makedirs(sample_out_dir, exist_ok=True)
    temp_unmapped_bam = os.path.join(sample_out_dir, f"{args.pid}_{sample_name}_unmapped.bam")
    print(f"Extracting unmapped reads from {bam_path} to {temp_unmapped_bam}")
    pysam.view('-b', '-f', '4', bam_path, '-o', temp_unmapped_bam, catch_stdout=False)
    pysam.index(temp_unmapped_bam)

    with pysam.AlignmentFile(temp_unmapped_bam, "rb") as unmapped_bam:
        unmapped_count = unmapped_bam.count(until_eof=True)
    print(f"Number of unmapped reads: {unmapped_count}")

    with pysam.AlignmentFile(bam_path, "rb" if bam_path.endswith('.bam') else "rc") as bam_file:
        total_count = bam_file.count(until_eof=True)
    print(f"Total number of reads in input: {total_count}")

    (_, read_lengths, _, _, _, _, repeat_threshold) = get_read_lengths_and_repeat_thresholds(args, None, bam_path)

    filter_telomere_reads.parallel_filter_telomere_reads(
        bam_path=temp_unmapped_bam,
        out_dir=sample_out_dir,
        pid=args.pid,
        sample=sample_name,
        repeat_threshold_calc=repeat_threshold,
        mapq_threshold=args.mapq_threshold,
        repeats=args.repeats,
        consecutive_flag=args.consecutive,
        remove_duplicates=args.remove_duplicates,
        band_file=band_file,
        num_processes=args.cores,
        singlecell_mode=getattr(args, "singlecell_mode", False),
        fast_mode=True,
    )
    print(f"Fast mode filtering complete for {sample_name}.")

    summary = create_fast_mode_sample_summary(args, sample_out_dir, bam_path, sample_name, total_count, read_lengths, repeat_threshold)
    return summary


def create_fast_mode_sample_summary(args, sample_out_dir, bam_path, sample_name, total_count, read_lengths, repeat_threshold):
    readcount_path = os.path.join(sample_out_dir, f"{args.pid}_readcount.tsv")
    filtered_bam_path = os.path.join(sample_out_dir, f"{args.pid}_filtered.bam")
    trpm_threshold = repeat_threshold if isinstance(repeat_threshold, int) else str(repeat_threshold)
    read_lengths_str = str(read_lengths) if read_lengths is not None else "NA"

    unmapped_reads = 0
    if os.path.exists(readcount_path):
        try:
            readcount_df = pd.read_csv(readcount_path, sep='\t')
            unmapped_reads = readcount_df[readcount_df['chr'] == 'unmapped']['reads'].sum()
        except Exception as e:
            print(f"Error reading readcount.tsv: {e}")

    tel_read_count = 0
    if os.path.exists(filtered_bam_path):
        try:
            with pysam.AlignmentFile(filtered_bam_path, "rb") as bam:
                tel_read_count = bam.count(until_eof=True)
        except Exception as e:
            print(f"Error reading filtered.bam for tel_read_count: {e}")

    TRPM = (tel_read_count / total_count * 1e6) if total_count > 0 else 0.0

    summary = {
        "PID": args.pid,
        "sample": sample_name,
        "total_reads": total_count,
        "read_length": read_lengths_str,
        "repeat_threshold_set": trpm_threshold,
        "repeat_threshold_used": trpm_threshold,
        "tel_read_count": tel_read_count,
        "TRPM": f"{TRPM:.6f}",
        "unmapped_reads": unmapped_reads
    }

    summary_path = os.path.join(sample_out_dir, f"{args.pid}_{sample_name}_summary.tsv")
    header = list(summary.keys())
    with open(summary_path, 'w') as f:
        f.write("\t".join(header) + "\n")
        f.write("\t".join(str(summary[h]) for h in header) + "\n")
    print(f"Summary file written for {sample_name}: {summary_path}")
    return summary


def write_fast_mode_combined_summary(args, out_dir, tumor_summary, control_summary):
    header = [
        "PID", "sample", "total_reads", "read_length", "repeat_threshold_set",
        "repeat_threshold_used", "tel_read_count", "TRPM", "unmapped_reads"
    ]
    rows = []
    tumor_trpm = None
    control_trpm = None

    def map_row(summary):
        return [str(summary.get(col, "NA")) for col in header]

    if tumor_summary:
        rows.append(map_row(tumor_summary))
        tumor_trpm = float(tumor_summary.get("TRPM", 0))
    if control_summary:
        rows.append(map_row(control_summary))
        control_trpm = float(control_summary.get("TRPM", 0))

    if tumor_trpm and control_trpm and control_trpm > 0 and tumor_trpm > 0:
        log2_trpm = np.log2(tumor_trpm / control_trpm)
        log2_row = [args.pid, "log2(tumor/control)"] + ["NA"] * (len(header) - 3)
        log2_row[header.index("TRPM")] = f"{log2_trpm:.6f}"
        log2_row.append("NA")
        rows.append(log2_row)

    combined_summary_path = os.path.join(out_dir, f"{args.pid}_summary.tsv")
    with open(combined_summary_path, 'w') as f:
        f.write("\t".join(header) + "\n")
        for row in rows:
            f.write("\t".join(row) + "\n")
    print(f"Combined summary file written: {combined_summary_path}")



#!/usr/bin/python

# Copyright 2024 Ferdinand Popp, Lina Sieverling, Philip Ginsbach, Lars Feuerbach

# This file is part of TelomereHunter2.

# TelomereHunter2 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# TelomereHunter2 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with TelomereHunter2. If not, see <http://www.gnu.org/licenses/>.

import os

import pysam

from telomerehunter2.utils import measure_time

#########################################################################
### gives the sum of all intratelomeric read lengths (base pairs)     ###
### secondary and supplementary alignments are skipped                ###
#########################################################################


@measure_time
def summed_intratelomeric_read_length(main_path, pid, sample):
    """
    Calculate the sum of intratelomeric read lengths and save the result in a TSV file.

    :param main_path: The main directory path.
    :param pid: Patient ID.
    :param sample: Sample ID.
    """

    # Output file path
    outfile_path = os.path.join(
        main_path, "TVRs", f"{pid}_{sample}_summed_read_length.tsv"
    )

    # Open output file for writing
    with open(outfile_path, "w") as outfile:
        # Write header
        outfile.write("PID\tsample\tsummed_intratel_read_length\n")

        # Input BAM file path
        bam_file_path = os.path.join(main_path, f"{pid}_filtered_intratelomeric.bam")

        # Open input BAM file for reading
        with pysam.AlignmentFile(bam_file_path, "rb") as bamfile:
            summed_read_length = 0

            # Iterate through reads
            for read in bamfile.fetch(until_eof=True):
                # Accumulate read length
                summed_read_length += len(read.query_sequence)

            # Write result to the output file
            outfile.write(f"{pid}\t{sample}\t{summed_read_length}\n")



#!/usr/bin/env python3
"""
sc-ATAC Barcode Splitter and Telomerehunter2 Runner
Splits sc-ATAC BAM files by cell barcodes (Sinto) and processes them with telomerehunter2
"""

import argparse
import datetime
import logging
import multiprocessing
import os
import re
import shutil
import subprocess
import sys
from collections import defaultdict
from concurrent.futures import ProcessPoolExecutor, as_completed
from pathlib import Path

import pysam

# Setup logging
logging.basicConfig(
    level=logging.INFO, format="%(asctime)s - %(levelname)s - %(message)s"
)
logger = logging.getLogger(__name__)


def sanitize_barcode(barcode):
    """Convert barcode to safe filename"""
    # Replace problematic characters with underscores
    safe_barcode = re.sub(r"[^a-zA-Z0-9_-]", "_", barcode)
    return safe_barcode


def get_barcodes_from_bam(
    input_bam,
    max_barcodes=None,
    barcode_txt_path=None,
    skip_barcode_file=False,
    min_reads_per_barcode=10000,
):
    logger.info(f"Barcodes from {input_bam}")

    barcode_counts = defaultdict(int)
    processed_reads = 0

    with pysam.AlignmentFile(input_bam, "rb") as bamfile:
        for read in bamfile.fetch(until_eof=True):
            processed_reads += 1
            if processed_reads % 100000 == 0:
                logger.info(
                    f"Processed {processed_reads} reads, found {len(barcode_counts)} barcodes"
                )
            bc = read.get_tag("CB") if read.has_tag("CB") else None
            if bc:
                barcode_counts[bc] += 1

    logger.info(
        f"Found: {len(barcode_counts)} unique barcodes in {processed_reads} reads"
    )

    # Filter barcodes by min_reads_per_barcode
    filtered_barcodes = {
        bc: count
        for bc, count in barcode_counts.items()
        if count >= min_reads_per_barcode
    }
    logger.info(
        f"{len(filtered_barcodes)} barcodes with at least {min_reads_per_barcode} reads"
    )

    # Optionally limit number of barcodes
    sorted_barcodes = sorted(
        filtered_barcodes.items(), key=lambda x: x[1], reverse=True
    )
    if max_barcodes and len(sorted_barcodes) > max_barcodes:
        logger.info(f"Limiting to top {max_barcodes} barcodes")
        sorted_barcodes = sorted_barcodes[:max_barcodes]

    # Write barcodes as: bc<TAB>bc<TAB>count
    if barcode_txt_path and not skip_barcode_file:
        with open(barcode_txt_path, "w") as f:
            for bc, count in sorted_barcodes:
                f.write(f"{bc}\t{bc}\t{count}\n")
        logger.info(f"Barcode list written to {barcode_txt_path}")
        logger.info(
            f"{len(sorted_barcodes)} barcodes written to file after filtering and limiting."
        )

    # Return dict: barcode -> read count
    return dict(sorted_barcodes)


def read_barcodes_from_file(barcode_txt_path, max_barcodes=None):
    """Read barcodes from an existing barcode file (strip fake group if present)"""
    logger.info(f"Reading barcodes from existing file: {barcode_txt_path}")
    barcodes = []
    with open(barcode_txt_path, "r") as f:
        for line in f:
            bc = line.strip()
            if bc:
                bc = bc.split("\t")[0]  # Remove fake group if present
                barcodes.append(bc)
    if max_barcodes and len(barcodes) > max_barcodes:
        logger.info(f"Limiting to top {max_barcodes} barcodes from file")
        barcodes = barcodes[:max_barcodes]
    logger.info(f"Loaded {len(barcodes)} barcodes from file")
    # No read counts available from file, so return as set
    return set(barcodes)


def split_with_sinto(
    input_bam, output_dir, patient_id, barcode_dict, barcode_txt_path, max_parallel=4
):
    """Use Sinto for efficient BAM splitting"""
    logger.info("Using Sinto for BAM splitting")

    # Prepare BAM output subfolder
    bam_out_dir = output_dir / "bam_files_by_barcode"
    bam_out_dir.mkdir(parents=True, exist_ok=True)

    # Check if BAM index exists, if not, create it
    bam_index = str(input_bam) + ".bai"
    if not os.path.exists(bam_index):
        logger.info(f"BAM index {bam_index} not found. Creating index with pysam...")
        pysam.index(str(input_bam))
        logger.info("BAM index created.")

    cmd = [
        "sinto",
        "filterbarcodes",
        "-b",
        str(input_bam),
        "-c",
        str(barcode_txt_path),
        "--outdir",
        str(bam_out_dir),
        "-p",
        str(max_parallel),
    ]

    logger.info(f"Running Sinto: {' '.join(cmd)}")

    try:
        result = subprocess.run(cmd, capture_output=True, text=True, check=True)
        logger.info("Sinto completed successfully")

        # Collect output files
        barcode_files = {}
        for barcode in barcode_dict:
            safe_bc = sanitize_barcode(barcode)
            bam_file = bam_out_dir / f"{safe_bc}.bam"

            if bam_file.exists():
                # Index BAM
                pysam.index(str(bam_file))
                # Rename to include patient ID
                new_name = bam_out_dir / f"{patient_id}_{safe_bc}.bam"
                shutil.move(str(bam_file), str(new_name))
                shutil.move(str(bam_file) + ".bai", str(new_name) + ".bai")
                barcode_files[barcode] = {
                    "file": str(new_name),
                    "safe_barcode": safe_bc,
                    "read_count": barcode_dict[barcode],
                }

        return barcode_files

    except subprocess.CalledProcessError as e:
        logger.error(f"Sinto failed: {e}")
        logger.error(f"Stdout: {e.stdout}")
        logger.error(f"Stderr: {e.stderr}")
        return None


def run_telomerehunter2(
    patient_barcode_id, bam_file, cytobanding_file, output_dir, th2_threads=1
):
    """Run telomerehunter2 on a single BAM file, mit Thread-Anzahl."""
    cmd = [
        "telomerehunter2",
        "-p",
        patient_barcode_id,
        "-ibt",
        str(bam_file),
        "-b",
        str(cytobanding_file),
        "-o",
        str(output_dir),
        "-pno",  # no plotting
        "-c",
        str(th2_threads),
    ]

    logger.info(
        f"Running telomerehunter2 for {patient_barcode_id} with {th2_threads} threads"
    )
    logger.debug(f"Command: {' '.join(cmd)}")

    env = os.environ.copy()
    env["OMP_NUM_THREADS"] = str(th2_threads)

    try:
        result = subprocess.run(
            cmd, cwd=output_dir, capture_output=True, text=True, check=True, env=env
        )
        logger.info(f"telomerehunter2 completed successfully for {patient_barcode_id}")
        return True, result.stdout, result.stderr
    except subprocess.TimeoutExpired:
        logger.error(f"telomerehunter2 timed out for {patient_barcode_id}")
        return False, "", "Process timed out"
    except subprocess.CalledProcessError as e:
        logger.error(
            f"telomerehunter2 failed for {patient_barcode_id} (exit code: {e.returncode})"
        )
        return False, e.stdout, e.stderr
    except Exception as e:
        logger.error(
            f"Unexpected error running telomerehunter2 for {patient_barcode_id}: {e}"
        )
        return False, "", str(e)


def th2_wrapper(args):
    """Wrapper for running telomerehunter2, for multiprocessing."""
    (
        barcode,
        file_info,
        patient_id,
        cytobanding_file,
        th2_results_dir,
        cleanup_bams,
        th2_threads,
    ) = args
    patient_barcode_id = f"{patient_id}_{file_info['safe_barcode']}"
    bam_file = file_info["file"]

    try:
        success, stdout, stderr = run_telomerehunter2(
            patient_barcode_id,
            bam_file,
            cytobanding_file,
            th2_results_dir,
            th2_threads=th2_threads,
        )
    except Exception as e:
        return barcode, {
            "success": False,
            "stdout": "",
            "stderr": str(e),
            "patient_barcode_id": patient_barcode_id,
        }

    # Clean up
    if cleanup_bams:
        try:
            os.unlink(bam_file)
            os.unlink(bam_file + ".bai")
        except Exception as e:
            logger.warning(f"Failed to clean up {bam_file}: {e}")

    return barcode, {
        "success": success,
        "stdout": stdout,
        "stderr": stderr,
        "patient_barcode_id": patient_barcode_id,
    }


def process_barcodes_with_telomerehunter2(
    barcode_files,
    patient_id,
    cytobanding_file,
    th2_results_dir,
    cleanup_bams=True,
    max_parallel=4,
):
    """Process all barcode BAM files with telomerehunter2, limiting parallelism and cores per job"""

    total_cores = multiprocessing.cpu_count()
    max_parallel = max_parallel if max_parallel < 6 else 6
    th2_threads = max(1, total_cores // max_parallel)
    logger.info(
        f"Processing {len(barcode_files)} barcodes with max {max_parallel} concurrent processes, each using {th2_threads} threads"
    )

    results = {}

    # Prepare argument tuples for each barcode
    args_list = [
        (
            barcode,
            file_info,
            patient_id,
            cytobanding_file,
            th2_results_dir,
            cleanup_bams,
            th2_threads,
        )
        for barcode, file_info in barcode_files.items()
    ]

    with ProcessPoolExecutor(max_workers=max_parallel) as executor:
        future_to_barcode = {
            executor.submit(th2_wrapper, args): args[0] for args in args_list
        }

        for i, future in enumerate(as_completed(future_to_barcode), 1):
            barcode = future_to_barcode[future]
            try:
                bc, result = future.result()
                results[bc] = result
                status = "SUCCESS" if result["success"] else "FAILED"
                logger.info(f"[{i}/{len(future_to_barcode)}] {bc} -> {status}")
            except Exception as e:
                logger.error(f"Exception processing {barcode}: {e}")
                results[barcode] = {"success": False, "error": str(e)}

    return results


def write_summary_report(
    results, patient_id, input_bam, th2_results_dir, cytobanding_file
):
    """Write detailed summary report"""

    summary_file = th2_results_dir / "telomerehunter2_summary.txt"

    successful = sum(1 for r in results.values() if r["success"])

    with open(summary_file, "w") as f:
        f.write("=== sc-ATAC Telomerehunter2 Processing Summary ===\n")
        f.write(f"Patient ID: {patient_id}\n")
        f.write(f"Input BAM: {input_bam}\n")
        f.write(f"Cytobanding file: {cytobanding_file}\n")
        f.write(f"Output directory: {th2_results_dir}\n")
        f.write(f"Total barcodes: {len(results)}\n")
        f.write(f"Successfully processed: {successful}\n")
        f.write(f"Failed: {len(results) - successful}\n")
        f.write(f"Success rate: {successful / len(results) * 100:.1f}%\n")
        f.write("\n=== Per-Barcode Results ===\n")
        f.write("Barcode\tStatus\tPatient_Barcode_ID\tError\n")

        for barcode, result in results.items():
            status = "SUCCESS" if result["success"] else "FAILED"
            error = (
                result.get("stderr", "").replace("\n", " ").strip()
                if not result["success"]
                else ""
            )
            patient_barcode_id = result.get("patient_barcode_id", "")
            f.write(f"{barcode}\t{status}\t{patient_barcode_id}\t{error}\n")

    logger.info(f"Summary report written to {summary_file}")
    return summary_file


def combine_summary_tsvs(output_dir, date_str):
    """
    Combine all *_summary.tsv files from telomerehunter2_results_YYYYMMDD/*/ subfolders into one file.
    """
    th2_results_dir = output_dir / f"telomerehunter2_results_{date_str}"
    combined_path = output_dir / f"combined_results_th2_{date_str}.tsv"

    # Look for summary files one level deeper: telomerehunter2_results_YYYYMMDD/*/*_summary.tsv
    summary_files = sorted(th2_results_dir.glob("*/*_summary.tsv"))
    if not summary_files:
        logger.warning(f"No summary TSV files found in {th2_results_dir} subfolders")
        return

    logger.info(
        f"Combining {len(summary_files)} summary TSV files into {combined_path}"
    )

    header_written = False
    with open(combined_path, "w") as outfile:
        for summary_file in summary_files:
            with open(summary_file, "r") as infile:
                lines = infile.readlines()
                if not lines:
                    continue
                # Write header only once
                if not header_written:
                    outfile.write(lines[0])
                    header_written = True
                # Write data lines (skip header)
                for line in lines[1:]:
                    outfile.write(line)
    logger.info(f"Combined summary written to {combined_path}")


def collect_existing_barcode_bams(bam_out_dir, patient_id, expected_barcodes=None):
    """
    Collects already split BAM files in the barcode folder and creates the barcode_files structure.
    Supports both {patient_id}_{barcode}.bam and {barcode}.bam file patterns.
    If expected_barcodes is given, checks that all are present and raises an error if not.
    """
    barcode_files = {}
    # Collect all BAM files matching both patterns
    all_bam_files = list(bam_out_dir.glob("*.bam"))
    for bam_file in all_bam_files:
        stem = bam_file.stem
        if stem.startswith(f"{patient_id}_"):
            safe_bc = stem[len(patient_id) + 1 :]
        else:
            safe_bc = stem
        barcode = (
            safe_bc  # Assumption: safe_barcode = barcode, adjust if mapping is needed
        )
        # Avoid duplicates (prefer patient_id_ prefix if both exist)
        if barcode not in barcode_files or stem.startswith(f"{patient_id}_"):
            barcode_files[barcode] = {
                "file": str(bam_file),
                "safe_barcode": safe_bc,
                "read_count": None,  # Read count unknown
            }
    # If expected_barcodes is provided, check completeness
    if expected_barcodes is not None:
        found = set(barcode_files.keys())
        expected = set(expected_barcodes)
        missing = expected - found
        logger.info(
            f"Found BAMs for {len(found & expected)}/{len(expected)} expected barcodes."
        )
        if missing:
            logger.error(
                f"Missing BAM files for {len(missing)} barcodes: {', '.join(sorted(missing))}"
            )
            raise RuntimeError(
                f"Not all expected barcode BAMs are present in {bam_out_dir}."
            )
    return barcode_files


def main():
    parser = argparse.ArgumentParser(
        description="Split sc-ATAC BAM by barcodes and run telomerehunter2",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog="""
Examples:
  # Run all steps (default)
  python %(prog)s input.bam ./output Patient1 cytobanding_hg38.txt

  # Only run telomerehunter2 on existing barcode BAMs
  python %(prog)s input.bam ./output Patient1 cytobanding_hg38.txt --steps th2

  # Only split and run th2 (use existing barcode file)
  python %(prog)s input.bam ./output Patient1 cytobanding_hg38.txt --steps split,th2

  # Only generate barcode file
  python %(prog)s input.bam ./output Patient1 cytobanding_hg38.txt --steps barcode
        """,
    )
    parser.add_argument(
        "-ibt", "--inputBamTumor", dest="input_bam", help="Input sc-ATAC BAM/CRAM file"
    )
    parser.add_argument("-o", "--output_dir", help="Output directory")
    parser.add_argument("-p", "--patient_id", help="Patient ID prefix")
    parser.add_argument(
        "-b", "--cytobanding_file", help="Cytobanding file for telomerehunter2"
    )
    parser.add_argument(
        "--max-barcodes",
        type=int,
        default=None,
        help="Maximum number of barcodes to process",
    )
    parser.add_argument(
        "--keep-bams", action="store_true", help="Keep individual barcode BAM files"
    )
    parser.add_argument(
        "--verbose", "-v", action="store_true", help="Enable verbose logging"
    )
    parser.add_argument(
        "--max-parallel",
        type=int,
        default=os.cpu_count(),
        help="Maximum number of parallel runs, default max number of cores",
    )
    parser.add_argument(
        "--min-reads-per-barcode",
        type=int,
        default=10000,
        help="Minimum number of reads per barcode to keep (default: 10000)",
    )
    parser.add_argument(
        "--steps",
        type=str,
        default="barcode,split,th2",
        help="Comma-separated list of steps to run: barcode,split,th2 (default: all)",
    )

    args = parser.parse_args()

    # Set logging level
    if args.verbose:
        logging.getLogger().setLevel(logging.DEBUG)

    # Print max_parallel and available cores
    available_cores = multiprocessing.cpu_count()
    logger.info(
        f"Max parallel: {args.max_parallel} | Available cores: {available_cores}"
    )

    # Parse steps
    steps = [s.strip().lower() for s in args.steps.split(",") if s.strip()]
    run_barcode = "barcode" in steps
    run_split = "split" in steps
    run_th2 = "th2" in steps

    # Validate inputs
    if not os.path.exists(args.input_bam):
        logger.error(f"Input BAM file not found: {args.input_bam}")
        sys.exit(1)

    if not os.path.exists(args.cytobanding_file):
        logger.error(f"Cytobanding file not found: {args.cytobanding_file}")
        sys.exit(1)

    # Create output directory
    output_dir = Path(args.output_dir)
    output_dir.mkdir(parents=True, exist_ok=True)
    barcode_txt_path = output_dir / "barcodes.txt"

    # Prepare subfolders
    bam_out_dir = output_dir / "bam_files_by_barcode"
    bam_out_dir.mkdir(parents=True, exist_ok=True)
    today_str = datetime.datetime.now().strftime("%Y%m%d")
    th2_results_dir = output_dir / f"telomerehunter2_results_{today_str}"
    th2_results_dir.mkdir(parents=True, exist_ok=True)

    # Step 1: Extract barcodes and write to text file
    if run_barcode:
        logger.info("=== Extracting barcodes and writing barcode list ===")
        barcode_dict = get_barcodes_from_bam(
            args.input_bam,
            args.max_barcodes,
            barcode_txt_path=barcode_txt_path,
            skip_barcode_file=False,
            min_reads_per_barcode=args.min_reads_per_barcode,
        )
        if not barcode_dict:
            logger.error("No barcodes found in input BAM")
            sys.exit(1)
        logger.info(f"{len(barcode_dict)} barcodes extracted")
        # If only barcode step is requested, exit after writing barcode file
        if steps == ["barcode"]:
            logger.info(
                "Only barcode step requested, exiting after barcode file generation."
            )
            sys.exit(0)
    else:
        if not barcode_txt_path.exists():
            logger.error(
                f"Barcode file {barcode_txt_path} does not exist but barcode step was skipped."
            )
            sys.exit(1)
        barcode_set = read_barcodes_from_file(barcode_txt_path, args.max_barcodes)
        if not barcode_set:
            logger.error("No barcodes found in existing barcode file")
            sys.exit(1)
        logger.info(f"{len(barcode_set)} barcodes loaded from existing barcode file")
        barcode_dict = {bc: None for bc in barcode_set}

    # Step 2: Sinto splitting
    if run_split:
        logger.info("=== Starting Sinto splitting ===")
        barcode_files = split_with_sinto(
            args.input_bam,
            output_dir,
            args.patient_id,
            barcode_dict,
            barcode_txt_path,
            max_parallel=args.max_parallel,
        )
        if not barcode_files:
            logger.error("Sinto splitting failed or no BAM files produced")
            sys.exit(1)
        logger.info(
            f"BAM successfully split into {len(barcode_files)} barcode-specific files"
        )
        # If only split step is requested, exit after splitting
        if steps == ["split"]:
            logger.info("Only split step requested, exiting after BAM splitting.")
            sys.exit(0)
    else:
        logger.info("Skipping BAM splitting, using existing barcode BAMs")
        if run_th2:
            try:
                barcode_files = collect_existing_barcode_bams(
                    bam_out_dir, args.patient_id, barcode_dict.keys()
                )
            except RuntimeError as e:
                logger.error(str(e))
                sys.exit(1)
            logger.info(f"{len(barcode_files)} barcode BAMs found")

    # Step 3: Telomerehunter2
    if run_th2:
        logger.info("=== Starting telomerehunter2 processing ===")
        results = process_barcodes_with_telomerehunter2(
            barcode_files,
            args.patient_id,
            args.cytobanding_file,
            th2_results_dir,
            cleanup_bams=not args.keep_bams,
            max_parallel=args.max_parallel,
        )
        summary_file = write_summary_report(
            results,
            args.patient_id,
            args.input_bam,
            th2_results_dir,
            args.cytobanding_file,
        )
        successful = sum(1 for r in results.values() if r["success"])
        logger.info("=== Processing finished ===")
        logger.info(f"Successfully processed: {successful}/{len(results)} barcodes")
        logger.info(f"Details: {summary_file}")
        if successful < len(results):
            logger.warning(f"{len(results) - successful} barcodes failed")
            logger.info("See summary file for error details")
        combine_summary_tsvs(output_dir, today_str)
        # If only th2 step is requested, exit after processing
        if steps == ["th2"]:
            logger.info(
                "Only th2 step requested, exiting after telomerehunter2 processing."
            )
            sys.exit(0)
    else:
        logger.info("Skipping telomerehunter2 step.")
        split_summary = output_dir / "split_summary.txt"
        with open(split_summary, "w") as f:
            f.write("=== BAM Splitting Summary ===\n")
            f.write(f"Input BAM: {args.input_bam}\n")
            f.write(f"Patient ID: {args.patient_id}\n")
            f.write(f"Total barcodes: {len(barcode_files)}\n")
            f.write(f"Output directory: {output_dir}\n\n")
            f.write("Barcode\tRead_Count\tOutput_File\n")
            for barcode, info in barcode_files.items():
                read_count = info.get("read_count", "NA")
                f.write(f"{barcode}\t{read_count}\t{info['file']}\n")
        logger.info(f"Split summary written to {split_summary}")


if __name__ == "__main__":
    main()



import os
import shutil
import subprocess
import sys
import tempfile
import unittest
from pathlib import Path

import pandas as pd
import pysam

class TestSCTelomereHunter2(unittest.TestCase):
    """Test cases for single-cell TelomereHunter2 summary script."""

    @classmethod
    def setUpClass(cls):
        cls.project_root = Path(__file__).resolve().parent.parent
        cls.test_dir = tempfile.mkdtemp(prefix="sc_telomerehunter2_test_")
        cls.test_dir = Path(cls.test_dir)
        cls.banding_file = (
            cls.project_root
            / "src"
            / "telomerehunter2"
            / "cytoband_files"
            / "hg19_cytoBand.txt"
        )
        if not cls.banding_file.exists():
            raise FileNotFoundError(f"Banding file not found at: {cls.banding_file}")

    @classmethod
    def tearDownClass(cls):
        shutil.rmtree(cls.test_dir)

    @classmethod
    def create_sc_bam_with_barcodes(cls, output_dir):
        bam_path = os.path.join(output_dir, "sc_test.bam")
        header = {
            "HD": {"VN": "1.6"},
            "SQ": [
                {"LN": 248956422, "SN": "chr1"},
                {"LN": 242193529, "SN": "chr2"},
            ],
        }
        with pysam.AlignmentFile(bam_path, "wb", header=header) as bam:
            for i in range(1, 6):
                read = pysam.AlignedSegment()
                read.query_name = f"sc_read_{i:03d}"
                read.query_sequence = "TTAGGG" * (i + 1)
                read.flag = 0
                read.reference_id = 0
                read.reference_start = 100 + i * 10
                read.mapping_quality = 60
                read.cigar = [(0, len(read.query_sequence))]
                read.next_reference_id = -1
                read.query_qualities = pysam.qualitystring_to_array("E" * len(read.query_sequence))
                read.set_tag("CB", f"CELL{i}")
                bam.write(read)
        pysam.sort("-o", bam_path.replace(".bam", ".sorted.bam"), bam_path)
        pysam.index(bam_path.replace(".bam", ".sorted.bam"))
        return Path(bam_path.replace(".bam", ".sorted.bam"))

    def test_sc_summary_script(self):
        results_path = self.test_dir / "sc_case"
        results_path.mkdir(exist_ok=True)
        sc_bam_path = self.create_sc_bam_with_barcodes(results_path)
        patient_name = "SC_TEST_PATIENT"
        sc_script = self.project_root / "src" / "telomerehunter2" / "telomerehunter2_sc.py"
        command = [
            sys.executable,
            str(sc_script),
            "-ibt", str(sc_bam_path),
            "-o", str(results_path),
            "-p", patient_name,
            "-b", str(self.banding_file),
            "--min-reads-per-barcode", "1"
        ]
        proc = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, universal_newlines=True)
        for line in proc.stdout:
            print(line, end="")
        exit_code = proc.wait()
        if exit_code != 0:
            error_message = proc.stderr.read()
            raise RuntimeError(f"SC summary script failed with exit code {exit_code}. Error: {error_message}")
        summary_file = results_path / patient_name / f"{patient_name}_sc_summary.tsv"
        expected_file = self.project_root / "tests" / "sc_case_expected_result.tsv"
        self.assertTrue(summary_file.exists(), f"SC summary file not found: {summary_file}")
        df = pd.read_csv(summary_file, sep="\t")
        self.assertIn("tel_content", df.columns, f"'tel_content' column not found in SC summary file: {summary_file}")

        # Check TVR column order: all TVRs arbitrary, then all TVRs 100bp, then all TVRs singletons
        TVR_HEXAMERS = [
            "TCAGGG", "TGAGGG", "TTGGGG", "TTCGGG", "TTTGGG",
            "ATAGGG", "CATGGG", "CTAGGG", "GTAGGG", "TAAGGG"
        ]
        tvr_arbitrary = [f"{tvr}_arbitrary_context_norm_by_intratel_reads" for tvr in TVR_HEXAMERS]
        tvr_100bp = [f"{tvr}_arbitrary_context_per_100bp_intratel_read" for tvr in TVR_HEXAMERS]
        tvr_singletons = [f"{tvr}_singletons_norm_by_all_reads" for tvr in TVR_HEXAMERS]
        std_columns = [
            'PID', 'sample', 'total_reads', 'read_length', 'repeat_threshold_set', 'repeat_threshold_used',
            'intratelomeric_reads', 'junctionspanning_reads', 'subtelomeric_reads', 'intrachromosomal_reads',
            'tel_read_count', 'gc_bins_for_correction', 'total_reads_with_tel_gc', 'tel_content', 'TRPM'
        ]
        expected_columns = std_columns + tvr_arbitrary + tvr_100bp + tvr_singletons
        self.assertEqual(list(df.columns), expected_columns, f"SC summary columns are not in the expected order.\nExpected: {expected_columns}\nFound: {list(df.columns)}")

        if not expected_file.exists():
            shutil.copy(summary_file, expected_file)
            df_expected = df
        else:
            df_expected = pd.read_csv(expected_file, sep="\t")
        pd.testing.assert_frame_equal(df, df_expected, check_dtype=False, check_index_type=False, check_column_type=False)
        print(f"✅ SC summary validation passed for {results_path}")

if __name__ == "__main__":
    unittest.main(verbosity=2)



#!/usr/bin/python

# Copyright 2024 Ferdinand Popp, Lina Sieverling, Philip Ginsbach, Lars Feuerbach

# This file is part of TelomereHunter2.

# TelomereHunter2 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# TelomereHunter2 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with TelomereHunter2. If not, see <http://www.gnu.org/licenses/>.


import argparse
import multiprocessing
import os
from concurrent.futures import ProcessPoolExecutor

import numpy as np
import pandas as pd
import pysam

from telomerehunter2 import (TVR_context, TVR_context_summary_tables,
                             TVR_screen, estimate_telomere_content,
                             filter_telomere_reads, get_repeat_threshold,
                             get_summed_intratelomeric_read_length, merge_pdfs,
                             normalize_TVR_counts, plot_functions,
                             repeat_frequency_intratelomeric,
                             sort_telomere_reads)
from telomerehunter2.utils import (
    file_exists,
    check_and_create_index,
    delete_temp_dir,
    check_and_prompt_filtering,
    get_repeat_threshold_from_summary,
    validate_args,
    validate_plotting_options,
    set_execution_flags,
    combine_summary_files,
    print_copyright_message,
    get_read_lengths_and_repeat_thresholds,
)
from telomerehunter2.fast_mode import run_fast_mode


def run_sample(
        sample_name,
        sample_bam,
        outdir_sample,
        sample_id,
        read_length,
        repeat_threshold_calc,
        repeat_threshold_str,
        filter_telomere_reads_flag,
        args,
):
    if getattr(args, "singlecell_mode", False):
        print(f"[TH2] Running sample '{sample_name}' in single-cell mode.")

    if filter_telomere_reads_flag:
        print(f"------ {sample_name}: started filtering telomere reads ------")
        filter_telomere_reads.parallel_filter_telomere_reads(
            bam_path=sample_bam,
            out_dir=outdir_sample,
            pid=args.pid,
            sample=sample_id,
            repeat_threshold_calc=repeat_threshold_calc,
            mapq_threshold=args.mapq_threshold,
            repeats=args.repeats,
            consecutive_flag=args.consecutive,
            remove_duplicates=args.remove_duplicates,
            band_file=args.banding_file,
            num_processes=args.cores,
            singlecell_mode=getattr(args, "singlecell_mode", False),
        )

    if args.sort_telomere_reads_flag:
        print("------ " + sample_name + ": started sorting telomere reads ------")
        sort_telomere_reads.sort_telomere_reads(
            input_dir=outdir_sample,
            out_dir=outdir_sample,
            band_file=args.banding_file,
            pid=args.pid,
            mapq_threshold=args.mapq_threshold,
            repeats=args.repeats,
        )

        # get a table with repeat frequencies per intratelomeric read
        repeat_frequency_intratelomeric.repeat_frequency_intratelomeric(
            input_path=outdir_sample,
            out_dir=outdir_sample,
            pid=args.pid,
            repeats=args.repeats,
        )

    if args.estimate_telomere_content_flag:
        print("------ " + sample_name + ": started estimating telomere content ------")

        estimate_telomere_content.get_gc_content_distribution(
            bam_file=os.path.join(
                outdir_sample, f"{args.pid}_filtered_intratelomeric.bam"
            ),
            out_dir=outdir_sample,
            pid=f"{args.pid}_intratelomeric",
            sample=sample_id,
            remove_duplicates=args.remove_duplicates,
        )

        estimate_telomere_content.estimate_telomere_content(
            input_dir=outdir_sample,
            out_dir=outdir_sample,
            sample=sample_id,
            read_length=read_length,
            repeat_threshold_str=repeat_threshold_str,
            pid=args.pid,
            repeat_threshold_set=args.repeat_threshold_set,
            per_read_length=args.per_read_length,
            gc_lower=args.gc_lower,
            gc_upper=args.gc_upper,
        )

    if args.TVR_screen_flag:
        print("------ " + sample_name + ": started TVR screen ------")

        # screen for TVRs
        TVR_screen.tvr_screen(
            main_path=outdir_sample,
            sample=sample_id,
            telomere_pattern="(.{3})(GGG)",
            min_base_quality=20,
            pid=args.pid,
            repeat_threshold_set=args.repeat_threshold_set,
        )

        # get summed intratelomeric read length
        get_summed_intratelomeric_read_length.summed_intratelomeric_read_length(
            main_path=outdir_sample, sample=sample_id, pid=args.pid
        )

    if args.TVR_context_flag:
        print("------ " + sample_name + ": started TVR context ------")

        # get context of selected TVRs
        # Todo not possible for flexible hexamers yet due to GGG and number of positions interesting
        for TVR in args.TVRs_for_context:
            TVR_context.tvr_context(
                main_path=outdir_sample,
                sample=sample_id,
                pattern=TVR,
                min_base_quality=20,
                telomere_pattern="GGG",
                tel_file="filtered_intratelomeric",
                pid=args.pid,
                context_before=args.bp_context,
                context_after=args.bp_context,
                repeat_threshold_set=args.repeat_threshold_set,
            )


def parse_command_line_arguments():
    # Cmd line input.
    parser = argparse.ArgumentParser(
        description="Estimation of telomere content from WGS data of a tumor and/or a control sample.",
        epilog="Contact Ferdinand Popp (f.popp@dkfz-heidelberg.de) for questions and support.",
    )

    input_files_group = parser.add_argument_group("Input Files")
    input_files_group.add_argument(
        "-p",
        "--pid",
        type=str,
        dest="pid",
        required=True,
        help="Sample name used in output files and diagrams (required).",
    )
    input_files_group.add_argument(
        "-o",
        "--outPath",
        type=str,
        dest="parent_outdir",
        required=True,
        help="Path to the output directory into which all results are written.",
    )
    input_files_group.add_argument(
        "-ibt",
        "--inputBamTumor",
        type=lambda x: file_exists(parser, x),
        dest="tumor_bam",
        help="Path to the indexed input BAM/CRAM file of the tumor sample.",
    )
    input_files_group.add_argument(
        "-ibc",
        "--inputBamControl",
        type=lambda x: file_exists(parser, x),
        dest="control_bam",
        help="Path to the indexed input BAM/CRAM file of the control sample.",
    )
    input_files_group.add_argument(
        "-b",
        "--bandingFile",
        type=lambda x: file_exists(parser, x),
        dest="banding_file",
        default=None,
        help="Optional: Path to a tab-separated file with information on chromosome banding. \
                                            The first four columns of the table have to contain the chromosome name, \
                                            the start and end position and the band name. The table should not have a header. \
                                            If no banding file is specified, scripts and plots will omit banding \
                                            information. Reference files are included for hg19 and hg38 in \
                                            telomerehunter2/cytoband_file. For example hg19_cytoBand.txt",
    )

    threshold_filtering_group = parser.add_argument_group(
        "Thresholds and Filtering Options"
    )
    threshold_filtering_group.add_argument(
        "-nf",
        "--noFiltering",
        dest="noFiltering",
        action="store_true",
        help="If the filtering step of TelomereHunter has already been run previously, skip this step.",
    )
    threshold_filtering_group.add_argument(
        "-rt",
        "--repeatThreshold",
        type=int,
        dest="repeat_threshold_set",
        help="The number of repeats needed for a read to be classified as telomeric. Minimum is 4. Values below 4 will be set to 4 automatically.",
    )
    threshold_filtering_group.add_argument(
        "-mqt",
        "--mappingQualityThreshold",
        type=int,
        dest="mapq_threshold",
        default=8,
        help="The mapping quality needed for a read to be considered as mapped (default = 8).",
    )
    threshold_filtering_group.add_argument(
        "-gc1",
        "--lowerGC",
        type=int,
        dest="gc_lower",
        default=48,
        help="Lower limit used for GC correction of telomere content.",
    )
    threshold_filtering_group.add_argument(
        "-gc2",
        "--upperGC",
        type=int,
        dest="gc_upper",
        default=52,
        help="Upper limit used for GC correction of telomere content.",
    )
    threshold_filtering_group.add_argument(
        "-con",
        "--consecutive",
        dest="consecutive",
        action="store_true",
        help="Search for consecutive repeats.",
    )
    threshold_filtering_group.add_argument(
        "-d",
        "--removeDuplicates",
        dest="remove_duplicates",
        action="store_true",
        help="Reads marked as duplicates in the input bam file(s) are removed in the filtering step.",
    )
    threshold_filtering_group.add_argument(
        "-rl",
        "--perReadLength",
        dest="per_read_length",
        action="store_true",
        help="Repeat threshold is set per 100 bp read length.",
    )
    threshold_filtering_group.add_argument(
        "--min-reads-per-barcode",
        type=int,
        dest="min_reads_per_barcode",
        default=10000,
        help="Minimum reads per barcode for single-cell summary only needed by telomerehunter2_sc.py not the bulk application (default: 10000).",
    )

    repeats_context_group = parser.add_argument_group("Repeats and Context Options")
    repeats_context_group.add_argument(
        "-r",
        "--repeats",
        nargs="+",
        dest="repeats",
        type=str,
        help="Base sequences to inspect like TTAGGG and its TVRs. First sequence is /"
             "used as base telomeric sequence. Default: ['TTAGGG', 'TGAGGG', 'TCAGGG', /"
             "'TTCGGG', 'TTGGGG'].",
        default=["TTAGGG", "TGAGGG", "TCAGGG", "TTGGGG", "TTCGGG", "TTTGGG"],
    )
    repeats_context_group.add_argument(
        "-rc",
        "--repeatsContext",
        nargs="+",
        dest="TVRs_for_context",
        type=str,
        default=[
            "TCAGGG",
            "TGAGGG",
            "TTGGGG",
            "TTCGGG",
            "TTTGGG",
            "ATAGGG",
            "CATGGG",
            "CTAGGG",
            "GTAGGG",
            "TAAGGG",
        ],
        help="List of telomere variant repeats for which to analyze the sequence context.",
    )
    repeats_context_group.add_argument(
        "-bp",
        "--bpContext",
        type=int,
        dest="bp_context",
        default=18,
        help="Number of base pairs on either side of the telomere variant repeat to investigate. Please use a number that is divisible by 6.",
    )

    running_group = parser.add_argument_group("Running Options")
    running_group.add_argument(
        "-c",
        "--cores",
        dest="cores",
        default=None,
        type=int,
        help="Specify the number of cores. If not provided, defaults to None and uses max cores.",
    )
    running_group.add_argument(
        "-pl",
        "--parallel",
        dest="run_parallel",
        action="store_true",
        help="The filtering, sorting, and estimating steps of the tumor and control sample are run in parallel. This will speed up the computation time of telomerehunter.",
    )
    running_group.add_argument(
        "--plot_mode",
        action="store_true",
        dest="plot_mode",
        help="Run only plotting steps on an existing analysis folder",
    )
    running_group.add_argument(
        "-sc",
        "--singlecell_mode",
        dest="singlecell_mode",
        action="store_true",
        help="Enable barcode counting/output for single-cell BAMs (CB tag). Default: auto-detect if CB tags present.",
    )
    running_group.add_argument(
        "-fm",
        "--fast_mode",
        dest="fast_mode",
        action="store_true",
        help="Run in fast mode: extract and filter unmapped reads only.",
    )

    plotting_group = parser.add_argument_group("Plotting Options")
    plotting_group.add_argument(
        "-pno",
        "--plotNone",
        dest="plotNone",
        action="store_true",
        default=False,
        help="Do not make any diagrams.",
    )
    plotting_group.add_argument(
        "-p1",
        "--plotChr",
        dest="plotChr",
        action="store_true",
        help="Make diagrams with telomeric reads mapping to each chromosome.",
    )
    plotting_group.add_argument(
        "-p2",
        "--plotFractions",
        dest="plotFractions",
        action="store_true",
        help="Make a diagram with telomere reads in each fraction.",
    )
    plotting_group.add_argument(
        "-p3",
        "--plotTelContent",
        dest="plotTelContent",
        action="store_true",
        help="Make a diagram with the gc corrected telomere content in the analyzed samples.",
    )
    plotting_group.add_argument(
        "-p4",
        "--plotGC",
        dest="plotGC",
        action="store_true",
        help="Make a diagram with GC content distributions in all reads and in intratelomeric reads.",
    )
    plotting_group.add_argument(
        "-p5",
        "--plotRepeatFreq",
        dest="plotRepeatFreq",
        action="store_true",
        help="Make histograms of the repeat frequencies per intratelomeric read.",
    )
    plotting_group.add_argument(
        "-p6",
        "--plotTVR",
        dest="plotTVR",
        action="store_true",
        help="Make plots for telomere variant repeats.",
    )
    plotting_group.add_argument(
        "-p7",
        "--plotSingleton",
        dest="plotSingleton",
        action="store_true",
        help="Make plots for singleton telomere variant repeats.",
    )
    plotting_group.add_argument(
        "-prc",
        "--plotRevCompl",
        dest="plotRevCompl",
        action="store_true",
        help="Distinguish between forward and reverse complement telomere repeats in diagrams.",
    )
    plotting_group.add_argument(
        "-pff",
        "--plotFileFormat",
        dest="plotFileFormat",
        default="pdf",
        choices=["pdf", "png", "svg", "all"],
        help="File format of output diagrams. Choose from pdf (default), png, svg or all (pdf, png and svg).",
    )

    # Create dict from parser args and save keys as variables
    args = parser.parse_args()

    # Set flags based on input arguments
    args.tumor_flag = bool(args.tumor_bam)
    args.control_flag = bool(args.control_bam)

    # Create output directory path
    args.outdir = os.path.join(args.parent_outdir, args.pid)

    # Validate arguments
    validate_args(args)
    validate_plotting_options(args)

    # Set execution flags
    set_execution_flags(args)

    # Input validation for repeat_threshold_set
    if args.repeat_threshold_set is not None and args.repeat_threshold_set < 4:
        print(
            "Warning: repeatThreshold was set below 4. Setting to minimum value of 4."
        )
        args.repeat_threshold_set = 4

    return args


def run_telomerehunter(
        args,
        control_bam,
        filter_telomere_reads_control,
        filter_telomere_reads_tumor,
        read_lengths_control,
        read_lengths_tumor,
        repeat_thresholds_control,
        repeat_thresholds_str_control,
        repeat_thresholds_str_tumor,
        repeat_thresholds_tumor,
        tumor_bam,
):
    # Determine maximum number of workers based on parallel execution setting
    max_workers = 1 if not args.run_parallel else min(2, multiprocessing.cpu_count())

    with ProcessPoolExecutor(max_workers=max_workers) as executor:
        submitted_futures = []

        # run tumor sample for PID
        if args.tumor_flag and (
                filter_telomere_reads_tumor
                or args.sort_telomere_reads_flag
                or args.estimate_telomere_content_flag
        ):
            tumor_sample_id = "tumor"
            tumor_output_dir = os.path.join(
                args.outdir, f"{tumor_sample_id}_TelomerCnt_{args.pid}"
            )
            os.makedirs(tumor_output_dir, exist_ok=True)

            # Submit to run_sample main function
            tumor_future = executor.submit(
                run_sample,
                sample_name="Tumor Sample",
                sample_bam=tumor_bam,
                outdir_sample=tumor_output_dir,
                sample_id=tumor_sample_id,
                read_length=read_lengths_tumor,
                repeat_threshold_calc=repeat_thresholds_tumor,
                repeat_threshold_str=repeat_thresholds_str_tumor,
                filter_telomere_reads_flag=filter_telomere_reads_tumor,
                args=args,
            )

            submitted_futures.append(tumor_future)

            if not args.run_parallel:
                tumor_future.result()

        # run control samples for PID
        if args.control_flag and (
                filter_telomere_reads_control
                or args.sort_telomere_reads_flag
                or args.estimate_telomere_content_flag
        ):
            control_sample_id = "control"
            control_output_dir = os.path.join(
                args.outdir, f"{control_sample_id}_TelomerCnt_{args.pid}"
            )
            os.makedirs(control_output_dir, exist_ok=True)

            # Submit to run_sample main function
            control_future = executor.submit(
                run_sample,
                sample_name="Control Sample",
                sample_bam=control_bam,
                outdir_sample=control_output_dir,
                sample_id=control_sample_id,
                read_length=read_lengths_control,
                repeat_threshold_calc=repeat_thresholds_control,
                repeat_threshold_str=repeat_thresholds_str_control,
                filter_telomere_reads_flag=filter_telomere_reads_control,
                args=args,
            )

            submitted_futures.append(control_future)

            if not args.run_parallel:
                control_future.result()

        # Wait for all futures to complete (only needed in parallel mode)
        if args.run_parallel:
            for future in submitted_futures:
                future.result()



def summary_log2(main_path, pid):
    # Construct the summary file path
    summary_file = os.path.join(main_path, f"{pid}_summary.tsv")

    # Read the summary table
    summary = pd.read_csv(summary_file, sep="\t", header=0)

    # Check which samples were run
    samples = summary["sample"]

    # If both "tumor" and "control" were run, calculate summary
    if "tumor" in samples.values and "control" in samples.values:
        start_log_index = 13  # tel_content and TRPM
        tumor_row = (
            summary.loc[samples == "tumor"].iloc[:, start_log_index:].fillna(0).values
        )
        control_row = (
            summary.loc[samples == "control"].iloc[:, start_log_index:].fillna(0).values
        )

        log2_values = np.empty_like(
            tumor_row, dtype=float
        )  # Creating an empty array to hold the result
        log2_values.fill(np.nan)
        mask = (control_row != 0.0) & (tumor_row != 0.0)
        log2_values[mask] = np.log2(np.divide(tumor_row[mask], control_row[mask]))

        # Replace inf and -inf with np.nan
        log2_values[np.isinf(log2_values)] = np.nan

        # Create a new row with [pid, "log2(tumor/control)"] and non-log empty columns
        new_row = pd.DataFrame(
            [[pid, "log2(tumor/control)"] + [np.nan] * (len(summary.columns) - 2)],
            columns=summary.columns,
        )

        # Add the calculated log2 values to the new row
        new_row.iloc[0, start_log_index:] = log2_values.flatten()

        # Append the new row to the summary table
        summary = pd.concat([summary, new_row], ignore_index=True)

    # convert cols to right type
    summary = summary.convert_dtypes()

    # Save extended summary table
    summary.to_csv(summary_file, sep="\t", index=False)


def run_plots(args, outdir, pid, repeat_thresholds_plot):
    if not args.plotNone:
        print("------ making plots ------\n")
        # Create plot directory
        os.makedirs(
            os.path.join(outdir, "plots"), exist_ok=True
        )  # all plots there except summary in outdir
        os.makedirs(
            os.path.join(outdir, "html_reports"), exist_ok=True
        )  # plotly folder

        if args.plotChr:
            print("Plot spectrum (plotChr)")
            plot_functions.plot_spectrum(
                outdir,
                pid,
                repeat_thresholds_plot,
                args.consecutive,
                args.mapq_threshold,
                args.banding_file,
                args.plotRevCompl,
                args.plotFileFormat,
            )

        if args.plotFractions:
            print("Plot spectrum summary (plotFractions)")
            plot_functions.plot_spectrum_summary(
                outdir,
                pid,
                repeat_thresholds_plot,
                args.consecutive,
                args.mapq_threshold,
                args.banding_file,
                args.plotRevCompl,
                args.plotFileFormat,
            )

        if args.plotTelContent:
            print("Plot telomere content (plotTelContent)")
            plot_functions.plot_tel_content(
                outdir,
                pid,
                repeat_thresholds_plot,
                args.consecutive,
                args.mapq_threshold,
                args.plotRevCompl,
                args.plotFileFormat,
                args.gc_lower,
                args.gc_upper,
            )

        if args.plotGC:
            print("Plot GC content (plotGC)")
            plot_functions.plot_gc_content(
                outdir, pid, args.plotFileFormat, args.gc_lower, args.gc_upper
            )

        if args.plotRepeatFreq:
            print("Plot intratelomeric repeat frequency (plotRepeatFreq)")
            plot_functions.plot_repeat_frequency_intratelomeric(
                outdir,
                pid,
                repeat_thresholds_plot,
                args.consecutive,
                args.mapq_threshold,
                args.repeats,
                args.plotFileFormat,
            )

        if args.plotTVR:
            print("Plot TVRs (plotTVR)")
            plot_functions.plot_TVR_plot(outdir, pid, args.plotFileFormat)

        if args.plotSingleton:
            print("Plot singletons (plotSingleton)")
            plot_functions.plot_singletons(outdir, pid, args.plotFileFormat)

        # Open to change due to plotly solutions
        # merge PDF plots if PDF files are created
        print("Combining results")
        merge_pdfs.merge_telomere_hunter_results(pid, outdir, args.banding_file)


def main():
    # print welcome
    print_copyright_message()

    # parse and check input parameters
    args = parse_command_line_arguments()
    if getattr(args, "singlecell_mode", False):
        print("Single-cell mode detected. Running full file analysis and barcode-specific insights.")

    # Fast mode branch
    if getattr(args, "fast_mode", False):
        run_fast_mode(args)
        return

    # If plot_mode is True, directly run plots and exit
    if args.plot_mode:
        repeat_thresholds_plot = get_repeat_threshold_from_summary(args)

        run_plots(args, args.outdir, args.pid, repeat_thresholds_plot)
        print("Completed plotting.")
        return

    # check if index is present, create it and if not writeable dir then do it in temp dir
    tumor_bam = (
        check_and_create_index(args.tumor_bam) if args.tumor_bam is not None else None
    )
    control_bam = (
        check_and_create_index(args.control_bam)
        if args.control_bam is not None
        else None
    )

    ################################################
    ### check if filtering has already been done ###
    ################################################

    filter_telomere_reads_T, filter_telomere_reads_C = check_and_prompt_filtering(
        args.filter_telomere_reads_flag,
        args.tumor_flag,
        args.control_flag,
        args.outdir,
        args.pid,
    )

    ###############################################
    #### Get Repeat Thresholds and Read Lengths ###
    ###############################################

    (
        read_lengths_control,
        read_lengths_tumor,
        repeat_thresholds_control,
        repeat_thresholds_plot,
        repeat_thresholds_str_control,
        repeat_thresholds_str_tumor,
        repeat_thresholds_tumor,
    ) = get_read_lengths_and_repeat_thresholds(args, control_bam, tumor_bam)

    ###############################
    ## run tumor sample for PID ###
    ###############################

    run_telomerehunter(
        args,
        control_bam,
        filter_telomere_reads_C,
        filter_telomere_reads_T,
        read_lengths_control,
        read_lengths_tumor,
        repeat_thresholds_control,
        repeat_thresholds_str_control,
        repeat_thresholds_str_tumor,
        repeat_thresholds_tumor,
        tumor_bam,
    )

    ##############
    ## Summary ###
    ##############

    # make a combined summary file of tumor and control results
    combine_summary_files(args.outdir, args.pid, args.tumor_flag, args.control_flag)

    # close possible temp dir
    delete_temp_dir(tumor_bam, control_bam, temp_dir="tmp_TH2")

    print("TVR_screen_flag")
    # make a table with normalized TVR counts
    if args.TVR_screen_flag:
        normalize_TVR_counts.normalize_TVR_counts(
            args.outdir, args.pid, ",".join(args.TVRs_for_context)
        )

    print("TVR_context_flag")
    # make TVR context summary tables
    if args.TVR_context_flag:
        TVR_context_summary_tables.tvr_context_singletons_tables(
            args.outdir,
            args.pid,
            args.repeats[0],
            args.bp_context,
            ",".join(args.TVRs_for_context),
        )

    # add log2 ratio to summary file
    if args.tumor_flag and args.control_flag:
        summary_log2(args.outdir, args.pid)

    ###############
    ## Plotting ###
    ###############

    run_plots(args, args.outdir, args.pid, repeat_thresholds_plot)

    print("Completed run.")


if __name__ == "__main__":
    main()



# __init__.py
from . import (TVR_context, TVR_screen, estimate_telomere_content,
               filter_telomere_reads, get_repeat_threshold,
               get_summed_intratelomeric_read_length, merge_pdfs,
               repeat_frequency_intratelomeric, sort_telomere_reads, utils)



#!/usr/bin/python

# Copyright 2024 Ferdinand Popp, Lina Sieverling, Philip Ginsbach, Lars Feuerbach

# This file is part of TelomereHunter2.

# TelomereHunter2 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# TelomereHunter2 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with TelomereHunter2. If not, see <http://www.gnu.org/licenses/>.

import cProfile
import io
import os
import pstats
import re
import shutil
import stat
import sys
import tempfile
import time

import pandas as pd
import pysam
from telomerehunter2 import get_repeat_threshold


def measure_time(func):
    """Wrapper function to simply print the execution time for a function"""

    def wrapper(*args, **kwargs):
        start_time = time.time()
        result = func(*args, **kwargs)
        end_time = time.time()
        elapsed_time = end_time - start_time
        print(f"Elapsed time for {func.__name__}: {elapsed_time} seconds")
        return result

    return wrapper


def profile_function(func, *args, **kwargs):
    """Wrapper function to profile a function"""
    pr = cProfile.Profile()
    pr.enable()
    result = func(*args, **kwargs)
    pr.disable()
    s = io.StringIO()
    sortby = "cumulative"
    ps = pstats.Stats(pr, stream=s).sort_stats(sortby)
    ps.print_stats(5)  # This will print only the top 10 entries
    print(s.getvalue())
    return result


def get_reverse_complement(dna_sequence):
    """Get the reverse complement of a DNA sequence."""
    complement_dict = {"A": "T", "T": "A", "C": "G", "G": "C", "N": "N"}
    complemented_sequence = (complement_dict[base] for base in reversed(dna_sequence))
    return "".join(complemented_sequence)


def get_band_info(banding_file):
    # Get band lengths
    band_info = pd.read_table(
        banding_file, sep="\t", names=["chr", "start", "end", "band_name", "stain"]
    )  # maybe not specify col names?
    band_info["chr"] = band_info["chr"].str.replace("chr", "")
    band_info["length"] = band_info["end"] - band_info["start"]
    chrs = band_info["chr"].unique()
    sorted_chromosomes = sorted(
        chrs, key=lambda x: (int(x) if x.isdigit() else float("inf"), x)
    )
    return band_info, sorted_chromosomes


def assure_dir_exists(path):
    dir = os.path.dirname(path)
    if not os.path.exists(dir):
        os.makedirs(dir)


def file_exists(parser, path):
    if path is None:
        return None
    if not os.path.exists(path):
        parser.error("The file %s does not exist!" % path)
    return path


def check_banding_file(banding_file, outdir=None):
    try:
        try:
            df = pd.read_csv(banding_file, sep="\t", header=None)
        except:
            df = pd.read_csv(banding_file, header=None)
            raise ValueError(f"Banding file is not a tsv file.")
        if df.shape[1] < 4:
            raise ValueError(
                f"File has fewer than 4 columns. Found {df.shape[1]} columns."
            )
        type_errors = []
        if not all(isinstance(x, str) for x in df[0].dropna()):
            type_errors.append("Column 1 should contain strings (chromosome names)")
        for col_idx in [1, 2]:
            df[col_idx] = pd.to_numeric(df[col_idx], errors="coerce")
            if df[col_idx].isna().any():
                type_errors.append(f"Column {col_idx + 1} should contain integers")
        if not all(isinstance(x, str) for x in df[3].dropna()):
            type_errors.append("Column 4 should contain strings (band names)")
        if type_errors:
            error_msg = "Invalid column types:\n" + "\n".join(
                f"- {err}" for err in type_errors
            )
            raise ValueError(error_msg)
        chromosome_level_only = df[0].nunique() == len(df)
        if chromosome_level_only:
            print("!!! Only chromosome-level data found, no band information. !!!")
            print("The banding file should be discarded.")
            subs_cytobands_df = generate_banding_file(df)
            print("\nGenerated substitution cytoband information (5% from chromosome ends):")
            print(subs_cytobands_df.head())
            print(f"Total generated bands: {len(subs_cytobands_df)}")
            generated_banding_file_path = (
                f"{outdir}/generated_bandingfile_5percent_ends.txt"
            )
            subs_cytobands_df.to_csv(generated_banding_file_path, sep="\t", index=False)
            print(
                f"Saved cytoband information (5% from chromosome ends): {generated_banding_file_path}"
            )
            return generated_banding_file_path
        else:
            print("Band-level information found. File is valid for detailed analysis.")
            return banding_file
    except Exception as e:
        print(f"ERROR: {str(e)}")
        sys.exit(1)


def generate_banding_file(df):
    bands_df = []
    for _, row in df.iterrows():
        chrom = row[0]
        start = row[1]
        end = row[2]
        band_width = end - start
        p_terminal = start + int(0.05 * band_width)
        q_terminal = end - int(0.05 * band_width)
        bands_df.append([chrom, start, p_terminal, f"{chrom}pter"])
        bands_df.append([chrom, p_terminal, q_terminal, f"{chrom}pq"])
        bands_df.append([chrom, q_terminal, end, f"{chrom}qter"])
    return pd.DataFrame(bands_df)


def get_bamfile_reference_genome(bam_file_path):
    try:
        mode = "rb" if bam_file_path.endswith(".bam") else "rc"
        with pysam.AlignmentFile(bam_file_path, mode=mode) as bam_file:
            header = bam_file.header.to_dict()
            reference_genome_string = header["SQ"][0]
            print(f"Input file is based on reference genome: {reference_genome_string}")
    except KeyError:
        return "Error when extracting reference genome from bam file."


def check_and_create_index(bam_path):
    bam_path = os.path.abspath(bam_path)
    index_path = bam_path + ".bai"
    if os.path.exists(index_path):
        return_bam = bam_path
    else:
        if os.access(os.path.dirname(bam_path), os.W_OK):
            pysam.index(bam_path)
            return_bam = bam_path
        else:
            temp_dir = os.path.abspath(tempfile.mkdtemp("tmp_TH2"))
            os.chmod(temp_dir, stat.S_IRUSR | stat.S_IWUSR | stat.S_IXUSR)
            temp_bam_path = os.path.join(temp_dir, os.path.basename(bam_path))
            try:
                os.symlink(bam_path, temp_bam_path)
            except OSError as e:
                print(f"Symbolic link creation failed: {e}")
            pysam.index(temp_bam_path)
            return_bam = temp_bam_path
    return return_bam


def delete_temp_dir(tumor_bam, control_bam, temp_dir):
    temp_exist = [
        os.path.dirname(path)
        for path in [tumor_bam, control_bam]
        if path is not None and temp_dir in path
    ]
    if temp_exist:
        try:
            shutil.rmtree(temp_exist[0])
            print(f"Folder '{temp_exist}' successfully deleted.")
        except Exception as e:
            print(f"Error deleting folder '{temp_exist}': {e}")


def has_filtering_output(outdir, pid, sample_id):
    sample_dir = os.path.join(outdir, f"{sample_id}_TelomerCnt_{pid}")
    files = [
        os.path.join(sample_dir, f"{pid}_filtered.bam"),
        os.path.join(sample_dir, f"{pid}_filtered_name_sorted.bam"),
        os.path.join(sample_dir, f"{pid}_readcount.tsv"),
        os.path.join(sample_dir, f"{pid}_{sample_id}_gc_content.tsv"),
    ]
    return all(os.path.exists(file) for file in files)


def check_and_prompt_filtering(
        filter_telomere_reads_flag, tumor_flag, control_flag, outdir, pid
):
    tumor_output_missing = (
        not has_filtering_output(outdir, pid, "tumor") if tumor_flag else False
    )
    control_output_missing = (
        not has_filtering_output(outdir, pid, "control") if control_flag else False
    )
    if tumor_output_missing and control_output_missing:
        print("Running filtering step as no output files are present")
        if filter_telomere_reads_flag:
            print(
                "!Denied skipping the filtering with --noFiltering, as no output files were present!"
            )
        filter_T, filter_C = True, True
    elif control_output_missing or tumor_output_missing:
        print("Running filtering step as not all output files are already present")
        if filter_telomere_reads_flag:
            print(
                "!Denied skipping the filtering with --noFiltering, as not all output files were present!"
            )
        filter_T, filter_C = tumor_output_missing, control_output_missing
    else:
        if filter_telomere_reads_flag:
            print(
                "Filtering samples, but output files are already present. Consider using the --noFiltering flag"
            )
            filter_T, filter_C = tumor_flag, control_flag
        else:
            print("Skipped filtering samples with --noFiltering flag")
            filter_T, filter_C = False, False
    return filter_T, filter_C


def get_repeat_threshold_from_summary(args):
    summary_file = os.path.join(args.outdir, f"{args.pid}_summary.tsv")
    try:
        summary = pd.read_csv(summary_file, sep="\t")
        if "repeat_threshold_used" in summary.columns:
            unique_thresholds = summary["repeat_threshold_used"].dropna().unique()
            if len(unique_thresholds) == 1:
                return str(int(unique_thresholds[0]))
            elif len(unique_thresholds) > 1:
                return "n"
        else:
            print(f"Error: Column 'repeat_threshold_used' not found in {summary_file}.")
            return "n"
    except Exception as e:
        print(f"Error reading file {summary_file}: {e}")
        return "n"


def validate_args(args):
    # check which bam files were specified
    if not args.plot_mode and not any([args.tumor_bam, args.control_bam]):
        raise ValueError(
            "argument -ibt/--inputBamTumor or -ibc/--inputBamControl or -icc or -ict is required"
        )

    if not args.tumor_bam:
        args.tumor_flag = False
        print(
            "Tumor BAM/CRAM file was not specified. Only running control BAM/CRAM file."
        )
    elif not args.control_bam:
        args.control_flag = False
        print(
            "Control BAM/CRAM file was not specified. Only running tumor BAM/CRAM file."
        )

    # Check band file format
    if args.banding_file:
        if args.tumor_bam or args.control_bam:
            get_bamfile_reference_genome(args.tumor_bam or args.control_bam)

        # check banding file and if only chromosomes, create 5% bands
        args.banding_file = check_banding_file(args.banding_file, args.outdir)
    else:
        print(
            "No banding file supplied, so all banding steps will be skipped and specific plots omitted."
        )

    # check if repeats only contains ACGT and are between 4 and 9 bases
    for repeat in args.repeats:
        if len(repeat) < 4 or len(repeat) > 9:
            raise ValueError(
                "argument -r/--repeats should be between 4 and 9 bases long."
            )
        x = re.search(r"[^ACGT]", repeat)
        if x is not None:
            raise ValueError(
                "argument -r/--repeats should only contain the letters ACGT."
            )

    for repeat in args.TVRs_for_context:
        x = re.search(r"[^ACGT]", repeat)
        if x is not None:
            raise ValueError(
                "argument -rc/--repeatsContext should only contain the letters ACGT."
            )

    # check if bp for sequencing context is divisible by 6 or base telomere sequence given
    if args.bp_context % len(args.repeats[0]) != 0:
        raise ValueError(
            f"argument -bp/--bpContext must be a multiple of length of first telomere repeat sequence /"
            f'{len(args.repeats[0])}. By default "TTAGGG" so multiple of 6 (e.g. 6, 12, 18, ...).'
        )

    if args.mapq_threshold < 0 or args.mapq_threshold > 40:
        raise ValueError(
            "argument -mqt/--mappingQualityThreshold must be an integer between 0 and 40."
        )

    if args.gc_lower < 0 or args.gc_lower > 100:
        raise ValueError(
            "argument -gc1/--lowerGC must be an integer between 0 and 100."
        )

    if args.gc_upper < 0 or args.gc_upper > 100:
        raise ValueError(
            "argument -gc2/--upperGC must be an integer between 0 and 100."
        )

    if args.gc_lower >= args.gc_upper:
        raise ValueError(
            "argument -gc1/--lowerGC must be less than argument -gc2/--upperGC."
        )


def validate_plotting_options(args):
    if args.plotNone and (
            args.plotChr
            or args.plotFractions
            or args.plotTelContent
            or args.plotGC
            or args.plotRepeatFreq
            or args.plotTVR
            or args.plotSingleton
    ):
        raise ValueError(
            "argument -pno/--plotNone should not be specified when other plotting options are selected."
        )

    # can be str or list so change to list
    if isinstance(args.plotFileFormat, str):
        args.plotFileFormat = [args.plotFileFormat]
    if args.plotFileFormat == "all":
        args.plotFileFormat = ["pdf", "png", "svg"]

    # if no plotting options are selected: plot all diagrams.
    if (
            not args.plotChr
            and not args.plotFractions
            and not args.plotTelContent
            and not args.plotGC
            and not args.plotRepeatFreq
            and not args.plotTVR
            and not args.plotSingleton
            and not args.plotNone
    ):
        args.plotChr = True
        args.plotFractions = True
        args.plotTelContent = True
        args.plotGC = True
        args.plotRepeatFreq = True
        args.plotTVR = True
        args.plotSingleton = True
    # no banding info remove plots
    if not args.banding_file:
        args.plotChr = False
        args.plotFractions = False


def set_execution_flags(args):
    args.filter_telomere_reads_flag = not args.noFiltering
    args.sort_telomere_reads_flag = args.estimate_telomere_content_flag = True
    args.TVR_screen_flag = args.TVR_context_flag = True

    if args.repeats[0] != "TTAGGG":
        print(
            "The first repeat to screen for is not canonical TTAGGG! Please double check."
        )
        args.TVR_screen_flag = args.TVR_context_flag = args.plotTVR = (
            args.plotSingleton
        ) = False


def combine_summary_files(outdir, pid, tumor_flag, control_flag):
    # Define the path for the combined summary file
    summary_path = os.path.join(outdir, f"{pid}_summary.tsv")

    tumor_summary_path = os.path.join(
        outdir, f"tumor_TelomerCnt_{pid}", f"{pid}_tumor_summary.tsv"
    )
    control_summary_path = os.path.join(
        outdir, f"control_TelomerCnt_{pid}", f"{pid}_control_summary.tsv"
    )

    # Check if both tumor and control flags are True
    if tumor_flag and control_flag:
        # Copy the tumor summary file to the combined summary file
        shutil.copyfile(tumor_summary_path, summary_path)

        # Append the last line of the control summary file to the combined summary file
        with open(control_summary_path, "r") as control_file, open(
                summary_path, "a"
        ) as combined_file:
            last_line = control_file.readlines()[-1]
            combined_file.write(last_line)

    # Check if only the tumor flag is True
    elif tumor_flag:
        # Copy the tumor summary file to the combined summary file
        shutil.copyfile(tumor_summary_path, summary_path)
    # Check if only the control flag is True
    elif control_flag:
        # Copy the control summary file to the combined summary file
        shutil.copyfile(control_summary_path, summary_path)


def print_copyright_message():
    print("\n")
    print(
        "\tTelomereHunter2 Copyright 2024 Ferdinand Popp, Lina Sieverling, Philip Ginsbach, Chen Hong, Lars Feuerbach"
    )
    print("\tThis program comes with ABSOLUTELY NO WARRANTY.")
    print("\tThis is free software, and you are welcome to redistribute it")
    print(
        "\tunder certain conditions. For details see the GNU General Public License v3.0"
    )
    print(
        "\tin the license copy received with TelomereHunter2 or <http://www.gnu.org/licenses/>."
    )
    print("\n")
    print("TelomereHunter2 1.0.0")
    print("\n")

def get_read_lengths_and_repeat_thresholds(args, control_bam, tumor_bam):
    """Calculate read lengths and repeat thresholds for tumor and control samples."""
    # Initialize all return variables
    read_lengths_str_control = None
    read_lengths_str_tumor = None
    repeat_thresholds_control = None
    repeat_thresholds_plot = None
    repeat_thresholds_str_control = None
    repeat_thresholds_str_tumor = None
    repeat_thresholds_tumor = None

    # Set default repeat threshold if not provided
    if not args.repeat_threshold_set:
        args.repeat_threshold_set = 6
        args.per_read_length = True
        print(
            "Repeat threshold per 100 bp was not set by the user. Setting it to 6 reads per 100 bp read length."
        )

    if args.per_read_length:
        # Calculate tumor thresholds if needed
        if args.tumor_flag:
            # Get read lengths and calculate thresholds for tumor
            read_lengths_str_tumor, tumor_read_length_counts = (
                get_repeat_threshold.get_read_lengths(tumor_bam)
            )
            print("Calculating repeat threshold for the tumor sample: ")
            repeat_thresholds_tumor, repeat_thresholds_str_tumor = (
                get_repeat_threshold.get_repeat_threshold(
                    read_lengths_str_tumor,
                    tumor_read_length_counts,
                    args.repeat_threshold_set,
                )
            )

        # Calculate control thresholds if needed
        if args.control_flag and control_bam:
            # Get read lengths and calculate thresholds for control
            read_lengths_str_control, control_read_length_counts = (
                get_repeat_threshold.get_read_lengths(control_bam)
            )
            print("Calculating repeat threshold for the control sample: ")
            repeat_thresholds_control, repeat_thresholds_str_control = (
                get_repeat_threshold.get_repeat_threshold(
                    read_lengths_str_control,
                    control_read_length_counts,
                    args.repeat_threshold_set,
                )
            )

        # Determine which threshold to use for plotting
        if args.tumor_flag and args.control_flag:
            repeat_thresholds_plot = (
                repeat_thresholds_tumor
                if repeat_thresholds_tumor == repeat_thresholds_control
                else "n"
            )
        elif args.tumor_flag:
            repeat_thresholds_plot = repeat_thresholds_tumor
        elif args.control_flag:
            repeat_thresholds_plot = repeat_thresholds_control

    else:
        # Use fixed threshold for all cases
        repeat_thresholds_tumor = args.repeat_threshold_set
        repeat_thresholds_control = args.repeat_threshold_set
        repeat_thresholds_plot = args.repeat_threshold_set
        repeat_thresholds_str_tumor = str(args.repeat_threshold_set)
        repeat_thresholds_str_control = str(args.repeat_threshold_set)

    print(
        f"Repeat Thresholds: Tumor={repeat_thresholds_tumor}, Control={repeat_thresholds_control}"
    )
    print("\n")

    return (
        read_lengths_str_control,
        read_lengths_str_tumor,
        repeat_thresholds_control,
        repeat_thresholds_plot,
        repeat_thresholds_str_control,
        repeat_thresholds_str_tumor,
        repeat_thresholds_tumor,
    )



#!/usr/bin/python

# Copyright 2024 Ferdinand Popp, Lina Sieverling, Philip Ginsbach, Lars Feuerbach

# This file is part of TelomereHunter2.

# TelomereHunter2 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# TelomereHunter2 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with TelomereHunter2. If not, see <http://www.gnu.org/licenses/>.


import os

from PyPDF2 import PdfMerger, PdfReader

from telomerehunter2.utils import get_band_info

###############################################################################
### merge all PDF plots produced as TelomereHunter output into one PDF file ###
###############################################################################


def merge_telomere_hunter_results(pid, outdir, banding_file=None):
    """
    Merge all TelomereHunter PDF plots into one PDF file.
    """
    # Define the possible PDF file names in correct order for merged PDF
    possible_file_names = [
        f"{pid}_telomere_content",
        f"{pid}_sorted_telomere_read_counts",
        f"{pid}_hist_telomere_repeats_per_intratelomeric_read",
        f"{pid}_gc_content",
        f"{pid}_TVR_barplot",
        f"{pid}_TVR_scatterplot",
        f"{pid}_TVR_plots_absolute_counts",
        f"{pid}_TVR_plots_log2_ratio",
        f"{pid}_pattern_screen_scatterplot",
        f"{pid}_singletons",
    ]

    if banding_file:
        _, chrs = get_band_info(banding_file)
        possible_file_names += [f"{pid}_{chromo}" for chromo in chrs]

    possible_pdf_names = [f"{name}.pdf" for name in possible_file_names]

    # Check which of the possible PDF files exist and sort
    files_dir = os.path.join(outdir, "plots")
    pdf_files = [f for f in os.listdir(files_dir) if f.endswith("pdf")]
    pdf_files_ordered = [pdf for pdf in possible_pdf_names if pdf in pdf_files]

    # Merge files
    merger = PdfMerger()

    for filename in pdf_files_ordered:
        pdf_path = os.path.join(files_dir, filename)
        merger.append(PdfReader(pdf_path, "rb"))

    # Write the merged PDF file
    merged_pdf_path = os.path.join(outdir, f"{pid}_all_plots_merged.pdf")
    merger.write(merged_pdf_path)

    print(f"Merged PDFs saved to: {merged_pdf_path}")

    # Merge all possible HTML files # maybe bring in order like pdfs
    possible_html_names = [f"{name}.html" for name in possible_file_names]
    html_dir = os.path.join(outdir, "html_reports")
    html_files = [f for f in os.listdir(html_dir) if f.endswith("html")]
    html_files_ordered = [html for html in possible_html_names if html in html_files]

    all_html_path = os.path.join(outdir, "all_reports.html")
    with open(all_html_path, "w") as output_html:
        for html_file in html_files_ordered:
            html_path = os.path.join(html_dir, html_file)
            with open(html_path, "r") as input_html:
                output_html.write(input_html.read())

    print(f"Merged HTML reports to: {all_html_path}")



#!/usr/bin/python

# Copyright 2024 Ferdinand Popp, Lina Sieverling, Philip Ginsbach, Lars Feuerbach

# This file is part of TelomereHunter2.

# TelomereHunter2 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# TelomereHunter2 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with TelomereHunter2. If not, see <http://www.gnu.org/licenses/>.

import os

import numpy as np
import pandas as pd


def process_pattern(
    main_path, pid, standard_repeat, bp_context, TVRs_for_context, summary
):
    df_list_all = {}
    multimer_repeats_count = bp_context // len(standard_repeat)

    for pattern in TVRs_for_context:
        # Initialize DataFrame with proper dtypes
        top_contexts = pd.DataFrame()

        for sample in summary["sample"].unique():
            TVR_context_dir = os.path.join(
                main_path, f"{sample}_TelomerCnt_{pid}", "TVR_context"
            )
            neighborhood_table_file = os.path.join(
                TVR_context_dir,
                f"{pid}_{sample}_{pattern}_{bp_context}bp_{bp_context}bp_neighborhood.tsv",
            )

            try:
                neighborhood_table = pd.read_table(neighborhood_table_file, sep="\t")
            except FileNotFoundError:
                print(f"Didnt find {neighborhood_table_file}")
                continue

            row_name = f"{pid}_{sample}_{bp_context}"

            # Basic fields
            top_contexts.loc[row_name, "PID"] = pid
            top_contexts.loc[row_name, "Sample"] = sample
            top_contexts.loc[row_name, "Context_bp"] = bp_context
            top_contexts.loc[row_name, "pattern"] = pattern

            # Handle empty neighborhood table
            if neighborhood_table.empty:
                print(f"{neighborhood_table_file} was empty")
                top_contexts.loc[row_name, "Bases"] = ""
                top_contexts.loc[row_name, "Count"] = 0
                top_contexts.loc[row_name, "Percent"] = 0.0
                top_contexts.loc[row_name, "Bases_t_type"] = ""
                top_contexts.loc[row_name, "Count_t_type"] = 0
                top_contexts.loc[row_name, "Percent_t_type"] = 0.0
                continue

            # Safe value assignments with type conversion
            top_contexts.loc[row_name, "Bases"] = str(
                neighborhood_table.loc[0, "Bases"]
            )
            top_contexts.loc[row_name, "Count"] = int(
                neighborhood_table.loc[0, "Count"]
            )
            top_contexts.loc[row_name, "Percent"] = float(
                neighborhood_table.loc[0, "Percent"]
            )

            # find singletons
            pattern_t_type_context = f"{standard_repeat * multimer_repeats_count}-{pattern}-{standard_repeat * multimer_repeats_count}"

            if pattern_t_type_context in neighborhood_table["Bases"].values:
                idx = neighborhood_table.index[
                    neighborhood_table["Bases"] == pattern_t_type_context
                ][0]
                top_contexts.loc[row_name, "Bases_t_type"] = str(
                    neighborhood_table.loc[idx, "Bases"]
                )
                top_contexts.loc[row_name, "Count_t_type"] = int(
                    neighborhood_table.loc[idx, "Count"]
                )
                top_contexts.loc[row_name, "Percent_t_type"] = float(
                    neighborhood_table.loc[idx, "Percent"]
                )
            else:
                top_contexts.loc[row_name, "Bases_t_type"] = ""
                top_contexts.loc[row_name, "Count_t_type"] = 0
                top_contexts.loc[row_name, "Percent_t_type"] = 0.0

            # Handle telomere content ratio
        if (
            "tumor" in summary["sample"].values
            and "control" in summary["sample"].values
        ):
            tel_content_tumor = summary.loc[
                summary["sample"] == "tumor", "tel_content"
            ].values[0]
            tel_content_control = summary.loc[
                summary["sample"] == "control", "tel_content"
            ].values[0]

            if tel_content_tumor == 0 or tel_content_control == 0:
                top_contexts["tel_content_log2_ratio"] = 0
            else:
                top_contexts["tel_content_log2_ratio"] = np.log2(
                    tel_content_tumor / tel_content_control
                )
        else:
            top_contexts["tel_content_log2_ratio"] = 0

        df_list_all[pattern] = top_contexts

    # Check if df_list_all is not empty before attempting to concatenate
    if df_list_all:
        pattern_contexts = pd.concat(df_list_all.values(), keys=df_list_all.keys())
    else:
        pattern_contexts = pd.DataFrame()  # Return an empty DataFrame
        print("Pattern context DF empty.")

    return pattern_contexts


def process_singleton_table(pattern_contexts, summary):
    singleton_table = pattern_contexts[
        ["PID", "pattern", "tel_content_log2_ratio", "Sample", "Count_t_type"]
    ]

    # Pivot the table using pivot_table to get Count_t_type per sample
    singleton_table = pd.pivot_table(
        singleton_table,
        values="Count_t_type",
        index=["PID", "pattern", "tel_content_log2_ratio"],
        columns="Sample",
        aggfunc="sum",
    )
    # Reset the index to make it flat
    singleton_table.reset_index(inplace=True)
    # Handle NaN values
    singleton_table.fillna(0, inplace=True)

    # Rename columns if 'control' and 'tumor' are present
    if "control" in singleton_table.columns:
        singleton_table.rename(
            columns={"control": "singleton_count_control"}, inplace=True
        )
        singleton_table["singleton_count_control"] = singleton_table[
            "singleton_count_control"
        ].astype(int)
    else:
        singleton_table["singleton_count_control"] = np.nan

    if "tumor" in singleton_table.columns:
        singleton_table.rename(columns={"tumor": "singleton_count_tumor"}, inplace=True)
        singleton_table["singleton_count_tumor"] = singleton_table[
            "singleton_count_tumor"
        ].astype(int)
    else:
        singleton_table["singleton_count_tumor"] = np.nan

    # Get log2 ratio of counts
    singleton_table["singleton_count_log2_ratio"] = np.log2(
        singleton_table["singleton_count_tumor"]
        / singleton_table["singleton_count_control"]
    )

    # Get total reads each for normalization
    total_reads_control = (
        summary.loc[summary["sample"] == "control", "total_reads"].values[0]
        if "control" in summary["sample"].values
        else 0
    )
    singleton_table["singleton_count_control_norm"] = (
        singleton_table["singleton_count_control"] / total_reads_control
    )

    total_reads_tumor = (
        summary.loc[summary["sample"] == "tumor", "total_reads"].values[0]
        if "tumor" in summary["sample"].values
        else 0
    )
    singleton_table["singleton_count_tumor_norm"] = (
        singleton_table["singleton_count_tumor"] / total_reads_tumor
    )

    singleton_table["singleton_count_log2_ratio_norm"] = np.log2(
        singleton_table["singleton_count_tumor_norm"]
        / singleton_table["singleton_count_control_norm"]
    )

    singleton_table["distance_to_expected_singleton_log2_ratio"] = (
        singleton_table["singleton_count_log2_ratio_norm"]
        - singleton_table["tel_content_log2_ratio"]
    )

    # move tel_ value to distance to make it better understandable
    tel_content_log2_ratio = singleton_table.pop("tel_content_log2_ratio")
    # Insert it at the second-to-last position
    singleton_table.insert(
        len(singleton_table.columns) - 1,
        "tel_content_log2_ratio",
        tel_content_log2_ratio,
    )

    return singleton_table


def add_tvr_normalized_counts(summary, TVRs_for_context, singleton_table):
    # Add column names
    singleton_colnames = [
        f"{pattern}_singletons_norm_by_all_reads" for pattern in TVRs_for_context
    ]
    for col in singleton_colnames:
        summary[col] = np.nan

    # Check if singleton_table is empty
    if singleton_table.empty:
        print(
            "Warning: singleton_table is empty. No TVR normalized counts will be added to the summary."
        )
        return summary, singleton_table

    # Set singleton table index to patterns
    singleton_table.index = singleton_table["pattern"]

    for sample_type in ["tumor", "control"]:
        if (singleton_table[f"singleton_count_{sample_type}"] != 0).any():
            mask = summary["sample"] == sample_type

            # Iterating through the present TVRs since only the ones found are in singleton table
            for pattern in TVRs_for_context:
                if pattern in singleton_table.index:
                    summary.loc[mask, f"{pattern}_singletons_norm_by_all_reads"] = (
                        singleton_table.loc[
                            pattern, f"singleton_count_{sample_type}_norm"
                        ]
                    )
        else:
            print(
                f"Warning: singleton_table {sample_type} has no non-zero counts. No TVR normalized counts will be added."
            )

    return summary, singleton_table


def save_results(summary, pattern_contexts, singleton_table, main_path, pid):
    try:
        summary.to_csv(
            os.path.join(main_path, f"{pid}_summary.tsv"), sep="\t", index=False
        )
    except Exception as e:
        print(f"Error saving extended summary table: {e}")

    if pattern_contexts.empty:
        print("Warning: pattern_contexts is empty. Saving an empty file.")

    patterns_top_contexts = pattern_contexts[
        ["PID", "Sample", "pattern", "Context_bp", "Bases", "Count", "Percent"]
    ]

    patterns_top_contexts.to_csv(
        os.path.join(main_path, f"{pid}_TVR_top_contexts.tsv"), sep="\t", index=False
    )

    if singleton_table.empty:
        print("Warning: singleton_table is empty. Saving an empty file.")
    singleton_table.to_csv(
        os.path.join(main_path, f"{pid}_singletons.tsv"), sep="\t", index=False
    )


def process_and_save_results(
    main_path, pid, summary, TVRs_for_context, pattern_contexts, singleton_table
):
    # add TVR counts to summary
    summary, singleton_table = add_tvr_normalized_counts(
        summary, TVRs_for_context, singleton_table
    )

    save_results(summary, pattern_contexts, singleton_table, main_path, pid)


def tvr_context_singletons_tables(
    main_path, pid, standard_repeat, bp_context, TVRs_for_context
):
    TVRs_for_context = TVRs_for_context.split(",")

    summary_file = os.path.join(main_path, f"{pid}_summary.tsv")
    summary = pd.read_table(summary_file, sep="\t")

    pattern_contexts = process_pattern(
        main_path, pid, standard_repeat, bp_context, TVRs_for_context, summary
    )

    if not pattern_contexts.empty and pattern_contexts["Count"].sum() > 0:
        singleton_table = process_singleton_table(pattern_contexts, summary)
    else:
        singleton_table = pd.DataFrame()

    process_and_save_results(
        main_path, pid, summary, TVRs_for_context, pattern_contexts, singleton_table
    )
